# coding: utf-8

"""
    unittest base
    ~~~~~~~~~~~~~

    :copyleft: 2009-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import os
import textwrap
import warnings

from django.test import TestCase
from django.urls import reverse

# https://github.com/jedie/django-tools
from django_tools.unittest_utils import assertments

from .BrowserDebug import debug_response


class BaseUnittestCase(TestCase):
    """
    Extensions to plain Unittest TestCase

    TODO: move all assert methods to django_tools/unittest_utils/assertments.py
    """
    maxDiff = 3000

    def _dedent(self, txt):
        # Remove any common leading whitespace from every line
        txt = textwrap.dedent(txt)

        # strip whitespace at the end of every line
        txt = "\n".join([line.rstrip() for line in txt.splitlines()])
        txt = txt.strip()
        return txt

    def assertEqual_dedent(self, first, second, msg=None):
        first = self._dedent(first)
        second = self._dedent(second)
        self.assertEqual(first, second, msg)

    def assertIn_dedent(self, member, container, msg=None):
        member = self._dedent(member)
        container = self._dedent(container)
        self.assertIn(member, container, msg)

    def assert_is_dir(self, path):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_is_dir!", DeprecationWarning)
        assertments.assert_is_dir(path)

    def assert_not_is_dir(self, path):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_path_not_exists!", DeprecationWarning)
        assertments.assert_path_not_exists(path)

    def assert_is_file(self, path):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_is_file!", DeprecationWarning)
        assertments.assert_is_file(path)

    def assert_not_is_File(self, path):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_path_not_exists!", DeprecationWarning)
        assertments.assert_path_not_exists(path)

    def assert_startswith(self, text, prefix):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_startswith!", DeprecationWarning)
        assertments.assert_startswith(text, prefix)

    def assert_endswith(self, text, prefix):
        warnings.warn("Use django_tools.unittest_utils.assertments.assert_endswith!", DeprecationWarning)
        assertments.assert_endswith(text, prefix)

    def assert_exception_startswith(self, context_manager, text):
        """
        e.g.:

        with self.assertRaises(AssertionError) as cm:
            do_something()

        self.assert_exception_startswith(cm, "First part of the error message")
        """
        exception_text = context_manager.exception.args[0]
        if not exception_text.startswith(text):
            msg = "%r doesn't starts with %r" % (exception_text, text)
            raise self.failureException(msg)

    def get_admin_url(self, obj, suffix):
        opts = obj._meta
        change_url = reverse(
            'admin:%s_%s_%s' % (opts.app_label, opts.model_name, suffix),
            args=(obj.pk,),
        )
        return change_url

    def get_admin_change_url(self, obj):
        """
        Get the admin change url for the given model instance.
        e.g.:
            "/admin/<app_name>/<model_name>/<pk>/"
        """
        return self.get_admin_url(obj, suffix="change")

    def get_admin_add_url(self, obj):
        """
        Get the admin add url for the given model.
        e.g.:
            "/admin/<app_name>/<model_name>/add/"
        """
        opts = obj._meta
        change_url = reverse('admin:%s_%s_add' % (opts.app_label, opts.model_name),)
        return change_url

    def get_messages(self, response):
        """
        Return all django message framwork entry as a normal list
        """
        return [str(message) for message in response.wsgi_request._messages]


class BaseTestCase(BaseUnittestCase):
    # Should we open a browser traceback?
    browser_traceback = True

    def raise_browser_traceback(self, response, msg):
        debug_response(response, self.browser_traceback, msg, display_tb=False)
        msg += ' (url: "%s")' % response.request.get("PATH_INFO", "???")
        raise self.failureException(msg)

    def assertStatusCode(self, response, excepted_code=200):
        """
        assert response status code, if wrong, do a browser traceback.
        """
        if response.status_code == excepted_code:
            return  # Status code is ok.
        msg = 'assertStatusCode error: "%s" != "%s"' % (response.status_code, excepted_code)
        self.raise_browser_traceback(response, msg)

    # def _assert_and_parse_html(self, html, user_msg, msg):
    #     """
    #     convert a html snippet into a DOM tree.
    #     raise error if snippet is no valid html.
    #     """
    #     try:
    #         return parse_html(html)
    #     except HTMLParseError as e:
    #         self.fail('html code is not valid: %s - code: "%s"' % (e, html))
    #
    # def _assert_and_parse_html_response(self, response):
    #     """
    #     convert html response content into a DOM tree.
    #     raise browser traceback, if content is no valid html.
    #     """
    #     try:
    #         return parse_html(response.content)
    #     except HTMLParseError as e:
    #         self.raise_browser_traceback(response, "Response's content is no valid html: %s' % e)

    def assertDOM(self, response, must_contain=(), must_not_contain=(), use_browser_traceback=True):
        """
        Asserts that html response contains 'must_contain' nodes, but no
        nodes from must_not_contain.
        """
        for txt in must_contain:
            try:
                self.assertContains(response, txt, html=True)
            except AssertionError as err:
                if use_browser_traceback:
                    self.raise_browser_traceback(response, err)
                raise

        for txt in must_not_contain:
            try:
                self.assertNotContains(response, txt, html=True)
            except AssertionError as err:
                if use_browser_traceback:
                    self.raise_browser_traceback(response, err)
                raise

    def assertMessages(self, response, messages):
        self.assertEqual(self.get_messages(response), messages)

    def assertResponse(
        self,
        response,
        must_contain=None,
        must_not_contain=None,
        status_code=200,
        template_name=None,
        messages=None,
        html=False,
        browser_traceback=True
    ):
        """
        Check the content of the response
        must_contain - a list with string how must be exists in the response.
        must_not_contain - a list of string how should not exists.
        """
        if must_contain is not None:
            for must_contain_snippet in must_contain:
                try:
                    self.assertContains(response, must_contain_snippet, status_code=status_code, html=html)
                except AssertionError as err:
                    if browser_traceback:
                        msg = 'Text not in response: "%s": %s' % (must_contain_snippet, err)
                        debug_response(response, self.browser_traceback, msg, display_tb=True)
                    raise

        if must_not_contain is not None:
            for must_not_contain_snippet in must_not_contain:
                try:
                    self.assertNotContains(response, must_not_contain_snippet, status_code=status_code, html=html)
                except AssertionError as err:
                    if browser_traceback:
                        msg = 'Text should not be in response: "%s": %s' % (must_not_contain_snippet, err)
                        debug_response(response, self.browser_traceback, msg, display_tb=True)
                    raise

        try:
            self.assertEqual(response.status_code, status_code)
        except AssertionError as err:
            if browser_traceback:
                msg = 'Wrong status code: %s' % err
                debug_response(response, self.browser_traceback, msg, display_tb=True)
            raise

        if template_name is not None:
            try:
                self.assertTemplateUsed(response, template_name=template_name)
            except AssertionError as err:
                if browser_traceback:
                    msg = 'Template not used: %s' % err
                    debug_response(response, self.browser_traceback, msg, display_tb=True)
                raise

        if messages is not None:
            try:
                self.assertMessages(response, messages)
            except AssertionError as err:
                if browser_traceback:
                    msg = 'Wrong messages: %s' % err
                    debug_response(response, self.browser_traceback, msg, display_tb=True)
                raise



"""
    :created: 28.08.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import unittest
from pathlib import Path

from django.test import SimpleTestCase

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.assertments import (
    assert_endswith, assert_installed_apps, assert_is_dir, assert_is_file, assert_language_code,
    assert_locmem_mail_backend, assert_path_not_exists, assert_startswith
)


class TestStringAssertments(unittest.TestCase):
    def test_startswith(self):
        assert_startswith("foobar", "foo")

    def test_not_startswith(self):

        with self.assertRaises(AssertionError) as cm:
            assert_startswith("foo", "bar")

        self.assertEqual(cm.exception.args[0], "'foo' doesn't starts with 'bar'")

    def test_endswith(self):
        assert_endswith("foobar", "bar")

    def test_not_endswith(self):

        with self.assertRaises(AssertionError) as cm:
            assert_endswith("foo", "bar")

        self.assertEqual(cm.exception.args[0], "'foo' doesn't ends with 'bar'")


class TestMailAssertments(SimpleTestCase):
    def test_assert_locmem_mail_backend(self):
        assert_locmem_mail_backend()


class TestLanguageCodeAssertments(unittest.TestCase):
    def test_assert_language_code(self):
        assert_language_code(language_code="de")
        assert_language_code(language_code="en")

    def test_assert_language_code_failed(self):
        self.assertRaises(AssertionError, assert_language_code, language_code="XX")


class TestInstalledAppsAssertments(unittest.TestCase):
    def test_assert_installed_apps(self):
        assert_installed_apps(app_names=["django.contrib.sessions", "django.contrib.admin"])

    def test_assert_installed_apps_failed(self):
        self.assertRaises(AssertionError, assert_installed_apps, app_names=("foo", "bar"))


class TestPathAssertments(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.existing_file_path = Path(__file__)
        assert cls.existing_file_path.is_file()
        cls.existing_dir_path = cls.existing_file_path.parent
        assert cls.existing_dir_path.is_dir()

    def test_is_existing_dir_with_path_instance(self):
        assert_is_dir(self.existing_dir_path)

    def test_is_existing_dir_with_string(self):
        assert_is_dir(str(self.existing_dir_path))

    def test_is_dir_failed(self):
        with self.assertRaises(AssertionError) as cm:
            assert_is_dir("/does/not/exists/")

        self.assertEqual(cm.exception.args[0], "Directory not exists: /does/not/exists")

    def test_is_existing_file_with_path_instance(self):
        assert_is_file(self.existing_file_path)

    def test_is_existing_file_with_string(self):
        assert_is_file(str(self.existing_file_path))

    def test_is_file_failed_not_directory(self):
        with self.assertRaises(AssertionError) as cm:
            assert_is_file("/path/to/does/not/exists/foo.txt")

        self.assertEqual(cm.exception.args[0], "Directory not exists: /path/to/does/not/exists")

    def test_is_file_failed_file_not_exists(self):
        with self.assertRaises(AssertionError) as cm:
            assert_is_file("/notexisting.txt")

        self.assertEqual(cm.exception.args[0], "File not exists: /notexisting.txt")

    def test_assert_path_not_exists(self):
        assert_path_not_exists("/path/to/does/not/exists/foo.txt")

    def test_assert_path_not_exists_is_file(self):
        with self.assertRaises(AssertionError) as cm:
            assert_path_not_exists(self.existing_file_path)

        self.assertEqual(cm.exception.args[0], "Path is a existing file: %s" % self.existing_file_path)

    def test_assert_path_not_exists_is_dir(self):
        with self.assertRaises(AssertionError) as cm:
            assert_path_not_exists(str(self.existing_dir_path))

        self.assertEqual(cm.exception.args[0], "Path is a existing directory: %s" % self.existing_dir_path)



# coding:utf-8

"""
    Dynamic SITE ID
    ~~~~~~~~~~~~~~~

    Set the SITE_ID dynamic by the current Domain Name.

    More info: read .../django_tools/dynamic_site/README.creole

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

import logging
import re

from django.contrib.sites import models as sites_models
from django.contrib.sites.models import Site
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from django_tools.local_sync_cache.local_sync_cache import LocalSyncCache
from django_tools.models import UpdateInfoBaseModel


logger = logging.getLogger("django_tools.DynamicSite")
if not logger.handlers:
    # ensures we don't get any 'No handlers could be found...' messages
    logger.addHandler(logging.NullHandler())


def compile_alias(alias):
    regex = re.compile(alias, re.UNICODE)
    return regex


class SiteAliasManager(models.Manager):
    ALIAS_CACHE = None

    def get_from_host(self, current_host):
        if self.ALIAS_CACHE is None:
            # first request after startup / model.save() -> fill the cache
            logger.debug("init LocalSyncCache for SiteAlias cache")
            self.ALIAS_CACHE = LocalSyncCache(id="SiteAliasCache")

        if "string" not in self.ALIAS_CACHE:
            logger.debug("Fill SiteAlias cache")
            self.ALIAS_CACHE["string"] = {}
            self.ALIAS_CACHE["regex"] = []
            queryset = self.all()
            for site_alias in queryset:
                site = site_alias.site
                alias = site_alias.alias

                if site_alias.regex:
                    regex = compile_alias(alias)
                    self.ALIAS_CACHE["regex"].append((regex, site))
                else:
                    self.ALIAS_CACHE["string"][alias.lower()] = site

            self.ALIAS_CACHE["regex"] = tuple(self.ALIAS_CACHE["regex"])
            logger.debug("Alias string cache: %s" % repr(self.ALIAS_CACHE["string"]))
            logger.debug("Alias regex cache: %s" % repr(self.ALIAS_CACHE["regex"]))

        # Try first all string alias:
        try:
            return self.ALIAS_CACHE["string"][current_host]
        except KeyError:
            logger.debug("No site found in string cache for %r" % current_host)

        # Try all regex alias:
        for regex, site in self.ALIAS_CACHE["regex"]:
            match = regex.search(current_host)
            if match is not None:
                return site

        logger.debug("No site found in regex cache for %r" % current_host)
        raise self.model.DoesNotExist("No alias found for %r" % current_host)

    def clear_cache(self):
        if self.ALIAS_CACHE is not None:
            logger.debug("Clear SiteAlias cache")
            self.ALIAS_CACHE.clear()
            logger.debug("Clear DynamicSiteMiddlewareCache cache")
            sites_models.SITE_CACHE.clear()


class SiteAlias(UpdateInfoBaseModel):
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    alias = models.CharField(max_length=256,
        help_text=_("A domain name alias for the site.")
    )
    regex = models.BooleanField(default=False,
        help_text=_("Is the given domain name alias a python regular expression?")
    )
    objects = SiteAliasManager()

    def clean_fields(self, exclude):
        message_dict = {}

        if "alias" not in exclude and "regex" not in exclude:
            if self.regex:
                alias = self.alias
                try:
                    compile_alias(alias)
                except Exception as err:
                    message_dict["alias"] = [mark_safe(_("Regex %r is not a valid: <strong>%s</strong>") % (alias, err))]

        if message_dict:
            raise ValidationError(message_dict)

    def save(self, *args, **kwargs):
        self.clean_fields(exclude={})
        UpdateInfoBaseModel.save(self, *args, **kwargs)
        SiteAlias.objects.clear_cache()

    class Meta:
        verbose_name = "site alias"
        verbose_name_plural = "site aliases"





# coding:utf-8

"""
    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import unicode_literals, absolute_import, print_function

from django.utils.html import strip_tags
from lxml.html.clean import Cleaner


def html2text(value):
    """
    Returns the text from HTML by strip all HTML tags.
    Used first Cleaner() from lxml and then django 'strip_tags'
    """
    if value:
        value = value.strip()

    if value:
        # use lxml cleaner:
        value = Cleaner().clean_html(value)

        # use Django
        value = strip_tags(value)

    return value



# coding: utf-8

"""
    Test /django_tools/management/commands/database_info.py

    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :created: 2017 by Jens Diemer
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import print_function, unicode_literals

import os
import sys
from unittest import TestCase

import django
from django.core.management import call_command
from django.test import SimpleTestCase

# https://github.com/jedie/django-tools
import django_tools
from django_tools.unittest_utils.django_command import DjangoCommandMixin
from django_tools.unittest_utils.stdout_redirect import StdoutStderrBuffer

MANAGE_DIR = os.path.abspath(os.path.join(os.path.dirname(django_tools.__file__), ".."))


class DatabaseInfoCallCommandTests(SimpleTestCase):
    """
    Test directly via django.core.management.call_command()
    Note:
        So the test database settings is active.
    """
    def test_database_info(self):
        with StdoutStderrBuffer() as buff:
            call_command("database_info")
        output = buff.get_output()
        print(output)

        self.assertIn("engine...............: 'sqlite3'", output)

        self.assertTrue(
            "name.................: ':memory:'" in output
            or
            "name.................: 'file:memorydb_default?mode=memory&cache=shared'" in output
        )

        self.assertIn("There are 1 connections.", output)

        self.assertIn("'CONN_MAX_AGE': 0,", output)


class DatabaseInfoSubprocessTests(DjangoCommandMixin, TestCase):
    """
    Test command via "manage.py" subprocess
    Note:
        So the real database settings is active and not the test settings!
    """
    def test_database_info(self):
        output = self.call_manage_py(["database_info"], manage_dir=MANAGE_DIR)
        print(output)
        self.assertIn("engine...............: 'sqlite3'", output)
        self.assertIn("name.................:", output)
        self.assertIn("test_project_db.sqlite3", output)

        self.assertIn("There are 1 connections.", output)

        self.assertIn("'CONN_MAX_AGE': 0,", output)



# coding: utf-8

from __future__ import unicode_literals

from copy import deepcopy

from django.conf import settings
from django.test import override_settings


TEMPLATE_INVALID_PREFIX='***invalid:'


class set_string_if_invalid(override_settings):
    """
    Decorator that set 'string_if_invalid' in settings.TEMPLATES
    more info in README

    see also:
    https://docs.djangoproject.com/en/1.8/ref/templates/api/#invalid-template-variables
    """
    def __init__(self):
        string_if_invalid='%s%%s***' % TEMPLATE_INVALID_PREFIX

        TEMPLATES = deepcopy(settings.TEMPLATES)
        for template_settings in TEMPLATES:
            template_settings['OPTIONS']['string_if_invalid'] = string_if_invalid

        super(set_string_if_invalid, self).__init__(TEMPLATES=TEMPLATES)



# coding: utf-8

"""
    Test DOM asserts
    ~~~~~~~~~~~~~~~~
    
    Test the DOM unittest stuff from
        Gregor Müllegger GSoC work in the Django soc2011/form-rendering branch

    https://github.com/gregmuellegger/django/blob/soc2011%2Fform-rendering/django/test/testcases.py
    https://github.com/gregmuellegger/django/blob/soc2011%2Fform-rendering/django/test/html.py
    
    :copyleft: 2011 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import unittest


if __name__ == "__main__":
    # run unittest directly
    import os
    os.environ["DJANGO_SETTINGS_MODULE"] = "django_tools.django_tools_tests.test_settings"


from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.selenium.response import FakedHttpResponse

FAKE_RESPONSE1 = FakedHttpResponse("""\
<!DOCTYPE HTML>
<html>
<head><title>A html5 page</title></head>
<body>
<form action="." method="post">
    <div style='display:none'><input type='hidden' name='csrfmiddlewaretoken' value='35e51f70b63058513d571a29465951b4' /></div>
    First name: <input type="text" name="first_name" value="Foo" /><br />
    Last name:<input type="text" name="last_name" value="Bar" /><br />
    <input type="submit" value="Submit" />
</form>
</body></html>""")


class DOMassertTest(BaseTestCase):

    def test_assertHTMLEqual1(self):
        self.assertHTMLEqual(
            '<a href="#" id="foo" class="bar">foobar</a>',
            '<a class="bar" id="foo" href="#">foobar</a>',
        )

    def test_assertHTMLEqual2(self):
        self.assertRaises(AssertionError,
            self.assertHTMLEqual,
            '<a href="#" id="foo" class="bar">foobar</a>',
            '<a class="bar" id="foo" href="#">other</a>',
        )

    def test_html_in_JavaScript1(self):
        """ TODO: https://github.com/gregmuellegger/django/issues/1 """
        self.assertHTMLEqual(
            '<script foo="1" bar="2">var js_sha_link="<p>***</p>"</script>',
            '<script bar="2" foo="1">var js_sha_link="<p>***</p>"</script>',
        )

    def test_html_in_JavaScript2(self):
        """ TODO: https://github.com/gregmuellegger/django/issues/1 """
        self.assertHTMLEqual(
            '<span foo="1" bar="2" /><script> <a href="" /> <p> <span></span> </p> </script>',
            '<span bar="2" foo="1" /><script> <a href="" /> <p> <span></span> </p> </script>',
        )

    def test_assertDOM1(self):
        self.assertDOM(
            response=FAKE_RESPONSE1,
            must_contain=(
                '<input name="last_name" type="text" value="Bar" />',
                '<input value="Foo" type="text" name="first_name" />',
            ),
            must_not_contain=("Traceback", "XXX INVALID TEMPLATE STRING"),
        )

    def test_assertDOM_must_contain(self):
        self.assertRaises(AssertionError,
            self.assertDOM,
            response=FAKE_RESPONSE1,
            must_contain=(
                '<input name="last_name" type="text" value="Bar" id="error" />',
            ),
            use_browser_traceback=False
        )

    def test_assertDOM_must_not_contain(self):
        self.assertRaises(AssertionError,
            self.assertDOM,
            response=FAKE_RESPONSE1,
            must_not_contain=(
                '<input name="last_name" type="text" value="Bar" />',
            ),
            use_browser_traceback=False
        )

    def test_assertContains1(self):
        self.assertContains(response=FAKE_RESPONSE1,
            text='<input name="last_name" type="text" value="Bar" />',
            count=1, html=True
        )

    def test_assertContains2(self):
        self.assertRaises(AssertionError,
            self.assertContains,
            response=FAKE_RESPONSE1,
            text='<input name="last_name" type="WRONG" value="Bar" />',
            count=1, html=True
        )


if __name__ == "__main__":
    # Run this unittest directly
    unittest.main()




"""
    :copyleft: 2017-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.core import mail
from django.core.mail import EmailMultiAlternatives
from django.test import SimpleTestCase

# https://github.com/jedie/django-tools
from django_tools.mail.send_mail import SendMail
from django_tools.unittest_utils.email import print_mailbox
from django_tools.unittest_utils.stdout_redirect import StdoutStderrBuffer
from django_tools.unittest_utils.unittest_base import BaseUnittestCase


class TestEMail(BaseUnittestCase, SimpleTestCase):

    def test_no_recipient(self):
        with self.assertRaises(AssertionError):
            SendMail(
                template_base="mail_test.{ext}",
                mail_context={
                    "foo": "first",
                    "bar": "second"
                },
                subject="Only a test",
                recipient_list=None
            )
        with self.assertRaises(AssertionError):
            SendMail(
                template_base="mail_test.{ext}",
                mail_context={
                    "foo": "first",
                    "bar": "second"
                },
                subject="Only a test",
                recipient_list=[]
            )

    def test_SendMail(self):
        self.assertEqual(len(mail.outbox), 0)

        ok = SendMail(
            template_base="mail_test.{ext}",
            mail_context={
                "foo": "first",
                "bar": "second"
            },
            subject="Only a test",
            recipient_list="foo@bar.tld"
        ).send()

        self.assertEqual(ok, True)

        print_mailbox(mail.outbox)

        self.assertEqual(len(mail.outbox), 1)
        email = mail.outbox[0]
        self.assertEqual(email.subject, 'Only a test')

        self.assertEqual_dedent(
            email.body, """
            <!-- START 'mail_test.txt' -->
            This is is a test mail.
            It used the django template: first, second
            
            <!-- END 'mail_test.txt' -->
        """
        )
        self.assertEqual(email.from_email, 'webmaster@localhost')
        self.assertEqual(email.to, ['foo@bar.tld'])

        self.assertIsInstance(email, EmailMultiAlternatives)

        html_email = email.alternatives[0]

        self.assertEqual_dedent(
            html_email[0].strip(), """
            <!-- START 'mail_test.html' -->
            <!DOCTYPE html>
            <html>
                <head></head>
                <body>
                    <p>This is is a test mail.</br>
                    It used the django template: first, second</p>
                </body>
            </html>
            
            
            <!-- END 'mail_test.html' -->
            """
        )
        self.assertEqual(html_email[1], 'text/html')



# coding: utf-8

"""
    :copyleft: 2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.conf import settings
from django.test import SimpleTestCase, override_settings
from django.core.mail.backends.locmem import EmailBackend
from django.utils import six

from django_tools.cache.site_cache_middleware import UpdateCacheMiddleware
from django_tools.utils.importlib import ImproperlyConfigured, get_attr_from_string, get_setting, \
    get_attr_from_settings, get_class_instance_from_settings

if six.PY3:
    from unittest import mock
else:
    import mock # https://pypi.python.org/pypi/mock


class TestImportlib(SimpleTestCase):
    def test_get_attr_from_string(self):
        UpdateCacheMiddleware2 = get_attr_from_string("django_tools.cache.site_cache_middleware.UpdateCacheMiddleware")
        self.assertEqual(UpdateCacheMiddleware2, UpdateCacheMiddleware)

        with self.assertRaises(ImproperlyConfigured) as cm:
            get_attr_from_string("no point", "TEST")
        self.assertEqual(cm.exception.args[0], "no point isn't a TEST module")

        with self.assertRaises(ImproperlyConfigured) as cm:
            get_attr_from_string("django.core.doesntexists", "Foo Bar")
        self.assertEqual(cm.exception.args[0], 'Foo Bar module "django.core" does not define a "doesntexists" class')

        with self.assertRaises(ImproperlyConfigured) as cm:
            get_attr_from_string("django_tools.utils.importlib.not_exists", "Import Lib")
        self.assertEqual(cm.exception.args[0], 'Import Lib module "django_tools.utils.importlib" does not define a "not_exists" class')

    def test_get_setting(self):
        self.assertEqual(
            get_setting("EMAIL_BACKEND"),
            'django.core.mail.backends.locmem.EmailBackend'
        )

    @mock.patch('django_tools.utils.importlib.logger')
    def test_get_setting_not_exists(self, mock_logger):
        get_setting("DOESN'T EXISTS")
        mock_logger.debug.assert_called_with(
            '''"DOESN'T EXISTS" not in settings defined'''
        )

    @mock.patch('django_tools.utils.importlib.logger')
    @override_settings(STATIC_URL=None)
    def test_get_setting_empty(self, mock_logger):
        self.assertEqual(settings.STATIC_URL, None)
        get_setting("STATIC_URL")
        mock_logger.debug.assert_called_with(
            'settings.STATIC_URL is None or empty'
        )

    def test_get_attr_from_settings(self):
        obj = get_attr_from_settings("EMAIL_BACKEND", "email backend")
        self.assertIs(obj, EmailBackend)

    @mock.patch('django_tools.utils.importlib.logger')
    def test_get_attr_from_settings_not_exists(self, mock_logger):
        get_attr_from_settings("DOESN'T EXISTS", "a test")
        mock_logger.debug.assert_called_with(
            '''"DOESN'T EXISTS" not in settings defined'''
        )

    @mock.patch('django_tools.utils.importlib.logger')
    @override_settings(STATIC_URL=None)
    def test_get_attr_from_settings_empty(self, mock_logger):
        self.assertEqual(settings.STATIC_URL, None)
        get_attr_from_settings("STATIC_URL", "a test")
        mock_logger.debug.assert_called_with(
            'settings.STATIC_URL is None or empty'
        )

    def test_get_class_instance_from_settings(self):
        obj = get_class_instance_from_settings("EMAIL_BACKEND", "email backend")
        self.assertIsInstance(obj, EmailBackend)

    @mock.patch('django_tools.utils.importlib.logger')
    def test_get_class_instance_from_settings_not_exists(self, mock_logger):
        get_class_instance_from_settings("DOESN'T EXISTS", "a test")
        mock_logger.debug.assert_called_with(
            '''"DOESN'T EXISTS" not in settings defined'''
        )

    @mock.patch('django_tools.utils.importlib.logger')
    @override_settings(STATIC_URL=None)
    def test_get_class_instance_from_settings_empty(self, mock_logger):
        self.assertEqual(settings.STATIC_URL, None)
        get_class_instance_from_settings("STATIC_URL", "a test")
        mock_logger.debug.assert_called_with(
            'settings.STATIC_URL is None or empty'
        )





"""
    :created: 24.04.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from pprint import pprint

from django.contrib.auth.models import User

from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME

from django.test import TestCase
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin


class ModelTestGeneratorAdminTestCase(TestUserMixin, BaseTestCase, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.create_testusers(cls)

    def test_generate_code_action_exists(self):
        self.login(usertype="superuser")

        response = self.client.get("/admin/auth/user/")

        self.assertResponse(response,
            must_contain=(
                '<strong>superuser</strong>',
                '<option value="generate_test_code">Generate unittest code</option>',
            ),
            must_not_contain=(
                "Error", "Traceback",
            ),
            status_code=200,
            template_name="admin/change_list.html",
            messages=[],
            html=False,
            browser_traceback=True
        )

    def test_generate_code(self):
        self.login(usertype="superuser")

        user_pks = User.objects.all().values_list("pk", flat=True)

        response = self.client.post(
            path="/admin/auth/user/",
            data={
                "action": "generate_test_code",
                ACTION_CHECKBOX_NAME: user_pks,
                "index":0,
            },
            HTTP_ACCEPT_LANGUAGE="en"
        )
        # TODO: check content-type and filename!
        self.assertResponse(response,
            must_contain=(
                "from auth.User <class 'django.contrib.auth.models.User'>",
                'user = User.objects.create(',

                "username='normal_test_user', # CharField, String (up to 150)",
                "username='staff_test_user', # CharField, String (up to 150)",
                "username='superuser', # CharField, String (up to 150)",
            ),
            must_not_contain=(
                "Error", "Traceback",
            ),
            status_code=200,
            template_name=None,
            messages=[],
            html=False,
            browser_traceback=True
        )
        headers = response._headers
        pprint(headers)
        self.assertEqual(headers["content-disposition"],
            ('Content-Disposition', 'attachment; filename=auth.User.py')
        )
        self.assertEqual(headers["content-type"], ('Content-Type', 'text/python'))



"""
    :created: 11.12.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import unittest

# https://github.com/jedie/django-tools
from django_tools.decorators import warn_class_usage, warn_function_usage


class TestWarnDecorators(unittest.TestCase):
    def test_warn_class_usage(self):
        @warn_class_usage(message="Test warn class usage !")
        class FooBar:
            pass

        with self.assertWarns(DeprecationWarning) as cm:
            FooBar()

        self.assertEqual(str(cm.warning), "Test warn class usage !")

    def test_warn_function_usage(self):
        @warn_function_usage(message="Test warn function usage !")
        def foo_bar():
            pass

        with self.assertWarns(DeprecationWarning) as cm:
            foo_bar()

        self.assertEqual(str(cm.warning), "Test warn function usage !")




"""
    :created: 24.04.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from django.contrib import messages
from django.http import HttpResponse
from django.utils.translation import ugettext_lazy as _

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.model_test_code_generator import ModelTestGenerator


def generate_test_code(modeladmin, request, queryset):
    """
    Django admin action to generate unittest code from model instances.

    To activate globally for all models, put this into one of your admin.py, e.g.:
        admin.site.add_action(generate_test_code)

    Select in admin some model entries and call the action from drop-down menu.
    """
    user = request.user
    if not user.is_superuser:
        messages.error(request, "Only allowed for superusers!")
        return

    model_test_generator = ModelTestGenerator()
    content = model_test_generator.from_queryset(queryset)
    # print(content)

    response = HttpResponse(content, content_type="text/python")

    model_label = modeladmin.model._meta.label
    response['Content-Disposition'] = 'attachment; filename=%s.py' % model_label

    return response


generate_test_code.short_description = _("Generate unittest code")



# coding: utf-8


"""
    directory selection
    ~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2011-2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import os

from django import forms
from django.db import models
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from django.utils.six import with_metaclass

from django_tools.utils.messages import failsafe_message
from django_tools import validators


class DirectoryWidget(forms.TextInput):
    """
    TODO: Add AJAX Stuff for easy select a existing directory path.
    """
    pass


class DirectoryFormField(forms.CharField):
    def __init__(self, base_path=settings.MEDIA_ROOT, *args, **kwargs):
        super(DirectoryFormField, self).__init__(*args, **kwargs)
        self.validators.append(
            validators.ExistingDirValidator(base_path=base_path)
        )

    def clean(self, value):
        value = super(DirectoryFormField, self).clean(value)
        value = os.path.normpath(value)
        return value


class DirectoryModelField(models.CharField):#, with_metaclass(models.SubfieldBase)):
    """
    >>> settings.DEBUG=False # Don't add path to error messages
    >>> dir = DirectoryModelField()
    >>> dir.run_validators(settings.MEDIA_ROOT)
    >>> dir.run_validators("does/not/exist")
    Traceback (most recent call last):
    ...
    django.core.exceptions.ValidationError: ["Directory doesn't exist!"]

    >>> dir.run_validators("../")
    Traceback (most recent call last):
    ...
    django.core.exceptions.ValidationError: ['Directory is not in base path!']

    >>> dir = DirectoryModelField(base_path="/")
    >>> dir.run_validators("/etc/default/")
    >>> dir.run_validators("var/log")

    >>> dir.run_validators("../bullshit")
    Traceback (most recent call last):
    ...
    django.core.exceptions.ValidationError: ["Directory doesn't exist!"]
    """
    default_validators = []
    description = _("A existing/accessible directory")

    def __init__(self, max_length=256, base_path=settings.MEDIA_ROOT, *args, **kwargs):
        super(DirectoryModelField, self).__init__(*args, max_length=max_length, **kwargs)
        self.validators.append(
            validators.ExistingDirValidator(base_path=base_path)
        )

    def formfield(self, **kwargs):
        """ Use always own widget and form field. """
        kwargs["widget"] = DirectoryWidget
        kwargs['form_class'] = DirectoryFormField
        return super(DirectoryModelField, self).formfield(**kwargs)



"""
    :created: 11.12.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import TestCase

# https://github.com/jedie/django-tools
from django_tools.selenium.chromedriver import SeleniumChromiumTestCase
from django_tools.selenium.geckodriver import SeleniumFirefoxTestCase


class SeleniumChromiumStaticLiveServerTestCase(TestCase, StaticLiveServerTestCase, SeleniumChromiumTestCase):
    """
    inherit only from 'LiveServerTestCase' will result in
    a empty database after test run.

    See:

    https://github.com/pytest-dev/pytest-django/issues/613
    https://code.djangoproject.com/ticket/25251
    https://github.com/django/django/pull/7528
    """

    pass


class SeleniumFirefoxStaticLiveServerTestCase(TestCase, StaticLiveServerTestCase, SeleniumFirefoxTestCase):
    """
    inherit only from 'LiveServerTestCase' will result in
    a empty database after test run.

    See:

    https://github.com/pytest-dev/pytest-django/issues/613
    https://code.djangoproject.com/ticket/25251
    https://github.com/django/django/pull/7528
    """

    pass



# coding: utf-8

"""
    Test django_tools.unittest_utils.django_command
"""

import os
import sys
from unittest import TestCase

# https://github.com/jedie/django-tools
import django_tools
from django_tools.unittest_utils.django_command import DjangoCommandMixin

PY35 = sys.version_info[0:2] == (3, 5)

MANAGE_DIR = os.path.abspath(os.path.join(os.path.dirname(django_tools.__file__), ".."))


class TestDjangoCommand(DjangoCommandMixin, TestCase):
    def test_help(self):
        output = self.call_manage_py(["--help"], manage_dir=MANAGE_DIR)

        self.assertNotIn("ERROR", output)
        self.assertIn("[django]", output)
        self.assertIn("Type 'manage.py help <subcommand>' for help on a specific subcommand.", output)

    def test_check(self):
        output = self.call_manage_py(["check"], manage_dir=MANAGE_DIR)
        self.assertNotIn("ERROR", output)
        self.assertIn("System check identified no issues (0 silenced).", output)

    def test_set_env(self):
        """
        Test if we can set "DJANGO_SETTINGS_MODULE"
        """
        env = dict(os.environ)
        env["DJANGO_SETTINGS_MODULE"]="does-not-exist"

        with self.assertRaises(AssertionError) as cm:
            self.call_manage_py(["--help"], manage_dir=MANAGE_DIR, env=env)

        output = "\n".join(cm.exception.args)
        # print(output)

        self.assertIn("subprocess exist status == 1", output)

        if PY35:
            member = "ImportError: No module named 'does-not-exist'"
        else:
            member = "ModuleNotFoundError: No module named 'does-not-exist'"

        self.assertIn(member, output)



# coding: utf-8

"""
    smooth cache backends
    ~~~~~~~~~~~~~~~~~~~~~

    more information in the README.

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

import logging
import os
import time

if __name__ == "__main__":
    # For doctest only
    os.environ["DJANGO_SETTINGS_MODULE"] = "django.conf.global_settings"

from django.conf import settings
from django.core.cache.backends.db import DatabaseCache
from django.core.cache.backends.filebased import FileBasedCache
from django.core.cache.backends.locmem import LocMemCache
from django.core.cache.backends.memcached import MemcachedCache, PyLibMCCache



logger = logging.getLogger("django-tools.SmoothCache")
if not logger.handlers:
    # ensures we don't get any 'No handlers could be found...' messages
    logger.addHandler(logging.NullHandler())

#if "runserver" in sys.argv or "tests" in sys.argv:
#    log.logging.basicConfig(format='%(created)f pid:%(process)d %(message)s')
#    logger.setLevel(log.logging.DEBUG)
#    logger.addHandler(log.logging.StreamHandler())


SMOOTH_CACHE_CHANGE_TIME = getattr(settings, "SMOOTH_CACHE_CHANGE_TIME", "DJANGO_TOOLS_SMOOTH_CACHE_CHANGE_TIME")
SMOOTH_CACHE_UPDATE_TIMESTAMP = getattr(settings, "SMOOTH_CACHE_UPDATE_TIMESTAMP", 10)
SMOOTH_CACHE_TIMES = getattr(settings, "SMOOTH_CACHE_TIMES", (
    # load value, max age in sec.
    (0, 5), #          < 0.1 ->  5sec
    (0.1, 10), #   0.1 - 0.5 -> 10sec
    (0.5, 30), #   0.5 - 1.0 -> 30sec
    (1.0, 60), #   1.0 - 1.5 ->  1Min
    (1.5, 120), #  1.5 - 2.0 ->  2Min
    (2.0, 300), #  2.0 - 3.0 ->  5Min
    (3.0, 900), #  3.0 - 4.0 -> 15Min
    (4.0, 3600), #     > 4.0 ->  1h
))

# Sort from biggest to lowest:
SMOOTH_CACHE_TIMES = list(SMOOTH_CACHE_TIMES)
SMOOTH_CACHE_TIMES.sort(reverse=True)
SMOOTH_CACHE_TIMES = tuple(SMOOTH_CACHE_TIMES)


def get_max_age(load_average):
    """
    return max age for the given load average.

    >>> get_max_age(0)
    5
    >>> get_max_age(0.09)
    5
    >>> get_max_age(0.1)
    5
    >>> get_max_age(0.11)
    10
    """
    for load, max_age in SMOOTH_CACHE_TIMES:
        if load_average > load:
            break
    return max_age


class SmoothCacheTime(int):
    def __new__(cls, value=None):
        if value is None:
            value = time.time()
        i = int.__new__(cls, value)
        return i


class _SmoothCache(object):
    __CHANGE_TIME = None # Timestamp of the "last update"
    __NEXT_SYNC = None # Point in the future to update the __CHANGE_TIME

    def __init__(self, *args, **kwargs):
        super(_SmoothCache, self).__init__(*args, **kwargs)
        self._smooth_clear = None

    def __get_change_time(self):
        """
        return current "last change" timestamp.
        To save cache access, update the timestamp only in
        SMOOTH_CACHE_UPDATE_TIMESTAMP frequency from cache.
        """
        now = time.time()
        if now > self.__NEXT_SYNC or self.__CHANGE_TIME is None:
            # Update timestamp we must look for a new change time from cache:
            self.__NEXT_SYNC = now + SMOOTH_CACHE_UPDATE_TIMESTAMP

            # use raw method, otherwise: end in a endless-loop ;)
            change_time = self.get(SMOOTH_CACHE_CHANGE_TIME, raw=True)
            if change_time is None:
                logger.debug("CHANGE_TIME is None")
                self.smooth_update() # save change time into cache
            elif change_time > self.__CHANGE_TIME:
                self.__CHANGE_TIME = change_time
                logger.debug("update change time to: %r" % change_time)
#        else:
#            logger.debug("Use old CHANGE_TIME %r" % self.__CHANGE_TIME)

        return self.__CHANGE_TIME

    def __must_updated(self, key, create_time):
        """
        return True if given cache create time is older than the "last change"
        time, but only if the additional time from system load allows it.
        """
        last_change_time = self.__get_change_time()
        if last_change_time < create_time:
            # Item was added to the cache after the last clear() time.
            logger.debug(
                "%r not out-dated: added %ssec before clear()" % (
                    key, (create_time - last_change_time)
            ))
            return False

        outdate_age = time.time() - last_change_time
        load_average = os.getloadavg()[0] # load over last minute
        max_age = get_max_age(load_average)
        if outdate_age > max_age:
            logger.debug("Out-dated %r (added %ssec after clear() - age: %s, max age: %s, load: %s)" % (
                key, (last_change_time - create_time), outdate_age, max_age, load_average
            ))
            return True
        else:
            logger.debug("Keep %r by load (out-dated age: %.1fsec, max age: %s, load: %s)" % (
                key, outdate_age, max_age, load_average
            ))
            return False

    def smooth_update(self):
        """
        save the "last change" timestamp to renew the cache entries in
        self.__CHANGE_TIME and in cache.
        """
        now = int(time.time())
        self.__CHANGE_TIME = now
        self.set(SMOOTH_CACHE_CHANGE_TIME, now, raw=True) # will be get via __origin_get()
        logger.debug("Set CHANGE_TIME to %r" % now)

    #--------------------------------------------------------------------------

    def get(self, key, default=None, version=None, raw=False):
        value = super(_SmoothCache, self).get(key, default, version)
        if raw:
            return value
        if value is None or value is default:
            # Item not in cache
            return value

        try:
            create_time, value = value
            assert isinstance(create_time, SmoothCacheTime), (
                "create_time is not SmoothCacheTime instance, it's: %s" % type(create_time)
            )
        except Exception as err:
            # e.g: entry is saved before smooth cache used.
            logger.error("Can't get 'create_time' from: %s (Maybe %r is a old cache entry?)" % (err, key))
            self.delete(key, version)
            return default

        if self.__must_updated(key, create_time):
            # is too old -> delete the item
            self.delete(key, version)
            return default

        return value

    def set(self, key, value, timeout=None, version=None, raw=False):
        if not raw:
            value = (SmoothCacheTime(), value)
        super(_SmoothCache, self).set(key, value, timeout, version)

    def clear(self):
        logger.debug("SmoothCache clear called!")
        super(_SmoothCache, self).clear()
        self.smooth_update()


class SmoothFileBasedCache(_SmoothCache, FileBasedCache):
    pass

class SmoothDatabaseCache(_SmoothCache, DatabaseCache):
    pass

class SmoothLocMemCache(_SmoothCache, LocMemCache):
    pass

class SmoothMemcachedCache(_SmoothCache, MemcachedCache):
    pass

class SmoothPyLibMCCache(_SmoothCache, PyLibMCCache):
    pass


if __name__ == "__main__":
    import doctest
    print(doctest.testmod())



# coding:utf-8

"""
    Client storage
    ~~~~~~~~~~~~~~
    
    Use dumps() and loads() from django.core.signing to store data into a Cookie.
    
    See:
    https://docs.djangoproject.com/en/1.4/topics/signing/#verifying-timestamped-values
    
    Usage e.g.:
    --------------------------------------------------------------------------
    from django_tools.utils.client_storage import SignedCookieStorageError, SignedCookieStorage
    
    def view1(request):
        response = HttpResponse("Hello World!")
        c = SignedCookieStorage(cookie_key="my_key", max_age=60)
        response = c.save_data(my_data, response)
        return response
    
    def view2(request):
        c = SignedCookieStorage(cookie_key="my_key", max_age=60)
        try:
            data = c.get_data(request)
        except SignedCookieStorageError, err:
            ...cookie missing or outdated or wrong data...
        else:
           ...do something with the data...
    --------------------------------------------------------------------------
    
    :copyleft: 2010-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

import os
import warnings

from django.core import signing


class SignedCookieStorageError(signing.BadSignature):
    pass


class SignedCookieStorage(object):
    """  
    see:
        django_tools_tests.test_signed_cookie.TestSignedCookieStorage
    """
    def __init__(self, cookie_key, max_age=60 * 60 * 24 * 7 * 52, compress=False):
        self.cookie_key = cookie_key
        self.max_age = max_age
        self.compress = compress

    def save_data(self, data, response, **kwargs):
        """ Add a cookie to response object, with data and security hash. """
        signed_data = signing.dumps(data, compress=self.compress, **kwargs)
        response.set_cookie(key=self.cookie_key, value=signed_data, max_age=self.max_age)
        return response

    def get_data(self, request):
        """ Return the stored data from cookie, if security hash is valid. """
        try:
            raw_data = request.COOKIES[self.cookie_key]
        except KeyError:
            raise SignedCookieStorageError("Cookie %r doesn't exists" % self.cookie_key)

        try:
            data = signing.loads(raw_data, max_age=self.max_age)
        except Exception as err:
            raise SignedCookieStorageError("Can't load data: %s" % err)

        return data



class ClientCookieStorage(object):
    """
    Support the old API

    TODO: remove in future
    """
    def __new__(self, *args, **kwargs):
        warnings.warn(
            "ClientCookieStorage is old API! Please change to SignedCookieStorage! This will be removed in the future!",
            FutureWarning,
            stacklevel=2
        )
        return SignedCookieStorage(*args, **kwargs)




# coding: utf-8

"""
    utilities for settings.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2015-2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from __future__ import absolute_import, division, print_function

from fnmatch import fnmatch


class IpPattern(str):
    def __init__(self, pattern):
        self.pattern = pattern

    def __eq__(self, pat):
        """
        ALLOWED_HOSTS compares via "<ip> == ALLOWED_HOSTS"
        see: django.http.request.validate_host
        """
        return fnmatch(pat, self.pattern)

    def lower(self):
        return IpPattern(self.pattern.lower())

    def startswith(self, *args):
        return False

    def __str__(self):
        return self.pattern


class FnMatchIps(list):
    """
    Allows you to use Unix shell-style wildcards of IP addresses
    for settings.INTERNAL_IPS and settings.ALLOWED_HOSTS
    used the fnmatch module.

    settings.py e.g.:
    --------------------------------------------------------------------------
    from django_tools.settings_utils import FnMatchIps

    INTERNAL_IPS = FnMatchIps(["127.0.0.1", "::1", "192.168.*.*", "10.0.*.*"])
    ALLOWED_HOSTS = FnMatchIps(["127.0.0.1", "::1", "192.168.*.*", "10.0.*.*"])
    --------------------------------------------------------------------------

    borrowed from https://djangosnippets.org/snippets/1380/
    """
    def __init__(self, pattern_list):
        super(FnMatchIps, self).__init__([IpPattern(pat) for pat in pattern_list])

    def __contains__(self, pat):
        # INTERNAL_IPS checks via "<ip> in INTERNAL_IPS"
        for ip in self:
            if pat == ip:
                return True
        return False


InternalIps = FnMatchIps # for compatibility



# coding: utf-8

"""
    Test limit to usergroups
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Unittests for limit_to_usergroups.py, see:
    https://github.com/jedie/django-tools/blob/master/django_tools/limit_to_usergroups.py

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

import unittest

from django.contrib.auth.models import AnonymousUser, Group
from django.core import management
from django.test import TestCase

from django_tools_test_project.django_tools_test_app.models import LimitToUsergroupsTestModel

# https://github.com/jedie/django-tools
from django_tools import limit_to_usergroups
from django_tools.limit_to_usergroups import get_verbose_limit_name, has_permission
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin

if __name__ == "__main__":
    # run unittest directly
    import os
    os.environ["DJANGO_SETTINGS_MODULE"] = "django_tools_test_project.test_settings"




class LimitToUsergroupsTest1(TestCase):
    def setUp(self):
        self.instance = LimitToUsergroupsTestModel.objects.create(
            permit_edit = limit_to_usergroups.UsergroupsModelField.SUPERUSERS,
            permit_view = limit_to_usergroups.UsergroupsModelField.ANONYMOUS_USERS,
        )

    def test_basic(self):
        self.assertEqual(
            self.instance.permit_edit,
            limit_to_usergroups.UsergroupsModelField.SUPERUSERS
        )
        self.assertEqual(
            self.instance.permit_view,
            limit_to_usergroups.UsergroupsModelField.ANONYMOUS_USERS
        )


class LimitToUsergroupsTest2(TestUserMixin, BaseTestCase, TestCase):
    def _pre_setup(self):
        super(LimitToUsergroupsTest2, self)._pre_setup()
        self.create_testusers()

    def setUp(self):
        self.user1 = self.login(usertype="staff")

        self.instance1 = LimitToUsergroupsTestModel.objects.create(
            permit_edit = limit_to_usergroups.UsergroupsModelField.STAFF_USERS,
            permit_view = limit_to_usergroups.UsergroupsModelField.ANONYMOUS_USERS,
        )
        self.user_group = Group.objects.create(name="user group 1")

        self.instance2 = LimitToUsergroupsTestModel.objects.create(
            permit_edit = self.user_group.pk,
            permit_view = limit_to_usergroups.UsergroupsModelField.NORMAL_USERS,
        )

    def test_filter_permission1(self):
        queryset = LimitToUsergroupsTestModel.objects.all()

        # Anonymous can view all
        items = limit_to_usergroups.filter_permission(
            queryset, permit_view=self.user1
        )
        self.assertEqual(len(items), 2)

    def test_filter_permission2(self):
        queryset = LimitToUsergroupsTestModel.objects.all()

        # edit only for staff or users in the right user group
        items = limit_to_usergroups.filter_permission(
            queryset, permit_edit=self.user1
        )
        self.assertEqual(len(items), 1)

    def test_filter_permission3(self):
        queryset = LimitToUsergroupsTestModel.objects.all()

        # Check if he is in the user group
        self.user1.groups.add(self.user_group)
        self.user1.save()
        items = limit_to_usergroups.filter_permission(
            queryset, permit_edit=self.user1
        )
        self.assertEqual(len(items), 2)

    def test_has_permission1(self):
        anonymous = AnonymousUser()

        self.assertTrue(
            has_permission(self.instance1, permit_view=anonymous)
        )
        self.assertFalse(
            has_permission(self.instance1, permit_edit=anonymous)
        )

    def test_has_permission2(self):
        anonymous = AnonymousUser()

        self.assertFalse(
            has_permission(self.instance2, permit_view=anonymous, permit_edit=anonymous)
        )

    def test_has_permission3(self):
        user = self.login(usertype="normal")
        self.assertFalse(
            has_permission(self.instance2, permit_view=user, permit_edit=user)
        )
        user.groups.add(self.user_group)
        user.save()
        self.assertTrue(
            has_permission(self.instance2, permit_view=user, permit_edit=user)
        )

    def test_verbose_limit_name1(self):
        self.assertEqual(
            get_verbose_limit_name(limit_to_usergroups.UsergroupsModelField.STAFF_USERS),
            'staff users'
        )

    def test_verbose_limit_name2(self):
        self.assertEqual(
            get_verbose_limit_name(self.user_group.pk),
            'user group 1'
        )


if __name__ == "__main__":
    # Run this unittest directly
    management.call_command('test',
        "django_tools.LimitToUsergroupsTest1",
        "django_tools.LimitToUsergroupsTest2",
        verbosity=2,
#        failfast=True
    )
    unittest.main()



"""
    :created: 2015 by Jens Diemer
    :copyleft: 2015-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging

from selenium import webdriver
from selenium.webdriver import DesiredCapabilities

# https://github.com/jedie/django-tools
from django_tools.selenium.base import SeleniumBaseTestCase
from django_tools.selenium.utils import find_executable

log = logging.getLogger(__name__)


class SeleniumChromiumTestCase(SeleniumBaseTestCase):
    """
    TestCase with Selenium and the Chromium WebDriver
    Note:
        * Needs 'chromium-chromedriver' executable! See README.creole for more info
        * It's without django StaticLiveServerTestCase

    usage e.g.:

        from django_tools.selenium.chromedriver import SeleniumChromiumTestCase, chromium_available

        @unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
        class ChromiumTests(SeleniumChromiumTestCase):

            def test_admin_login_page(self):
                self.driver.get(self.live_server_url + "/admin/login/")
                self.assert_equal_page_title("Log in | Django site admin")
                self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
                self.assert_no_javascript_alert()

    see also: django_tools_tests/test_unittest_selenium.py
    """

    filename = "chromedriver"

    # Overwrite this in sub class, if needed:
    extra_search_paths = ("/usr/lib/chromium-browser",)

    options = ("--no-sandbox", "--headless", "--disable-gpu")
    desired_capabilities = {
        "loggingPrefs": {"browser": "ALL", "client": "ALL", "driver": "ALL", "performance": "ALL", "server": "ALL"}
    }

    @classmethod
    def setUpClass(cls):

        chrome_options = webdriver.ChromeOptions()
        for argument in cls.options:
            chrome_options.add_argument(argument)

        try:
            executable = find_executable(cls.filename, cls.extra_search_paths)
        except FileNotFoundError:
            cls.driver = None
        else:
            desired = DesiredCapabilities.CHROME
            for key, value in cls.desired_capabilities.items():
                desired[key] = value

            cls.driver = webdriver.Chrome(
                chrome_options=chrome_options,
                executable_path=str(executable),  # Path() instance -> str()
                desired_capabilities=desired,
            )
            cls.driver.implicitly_wait(10)

        super().setUpClass()

    def get_browser_log(self):
        assert "browser" in self.driver.log_types
        browser_log = self.driver.get_log("browser")
        lines = []
        for entry in browser_log:
            lines.append(("{timestamp} {level} {source} {message}").format(**entry))
        return "\n".join(lines)

    def assert_in_browser_log(self, text):
        log = self.get_browser_log()
        print(log)
        self.assertIn(text, log)


def chromium_available(filename=None):
    """
    :return: True/False if 'chromium-chromedriver' executable can be found

    >>> chromium_available("doesn't exists")
    False
    """
    if filename is None:
        filename = SeleniumChromiumTestCase.filename

    try:
        executable = find_executable(filename, SeleniumChromiumTestCase.extra_search_paths)
    except FileNotFoundError as err:
        log.error("Chromium is no available: %s")
        return False

    log.debug("Chromium found here: %s" % executable)
    return True



"""
    :created: 28.08.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from pathlib import Path

from django.conf import settings
from django.core import mail
from django.core.mail import get_connection


def assert_startswith(text, prefix):
    """
    Check if test starts with prefix.
    """
    assert text.startswith(prefix), "%r doesn't starts with %r" % (text, prefix)


def assert_endswith(text, suffix):
    """
    Check if text ends with suffix.
    """
    assert text.endswith(suffix), "%r doesn't ends with %r" % (text, suffix)


def assert_locmem_mail_backend():
    """
    Check if current email backend is the In-memory backend.
    See:
        https://docs.djangoproject.com/en/1.11/topics/email/#in-memory-backend
    """
    mail_backend = get_connection()
    assert isinstance(mail_backend, mail.backends.locmem.EmailBackend), "Wrong backend: %s" % mail_backend


def assert_language_code(*, language_code):
    """
    Check if given language_code is in settings.LANGUAGES
    """
    existing_language_codes = tuple(dict(settings.LANGUAGES).keys())
    assert language_code in existing_language_codes, "%r not in settings.LANGUAGES=%r" % (
        language_code,
        settings.LANGUAGES,
    )


def assert_installed_apps(*, app_names):
    """
    Check entries in settings.INSTALLED_APPS
    """
    assert isinstance(app_names, (tuple, list))
    installed_apps = settings.INSTALLED_APPS
    for app_name in app_names:
        assert app_name in installed_apps, "%r not in settings.INSTALLED_APPS!" % app_name


def assert_is_dir(path):
    """
    Check if given path is a directory
    """
    if not isinstance(path, Path):
        path = Path(path)
    assert path.is_dir(), "Directory not exists: %s" % path


def assert_is_file(path):
    """
    Check if given path is a file
    """
    if not isinstance(path, Path):
        path = Path(path)
    assert_is_dir(path.parent)
    assert path.is_file(), "File not exists: %s" % path


def assert_path_not_exists(path):
    """
    Check if given path doesn't exists
    """
    if not isinstance(path, Path):
        path = Path(path)

    assert not path.is_dir(), "Path is a existing directory: %s" % path
    assert not path.is_file(), "Path is a existing file: %s" % path
    assert not path.is_fifo(), "Path is a existing fifo: %s" % path
    assert not path.exists(), "Path exists: %s" % path




import os

from django.http import Http404
from django.test import SimpleTestCase
from django.test.utils import override_settings
from django.utils import six

import django_tools
from django_tools.filemanager.exceptions import DirectoryTraversalAttack
from django_tools.filemanager.filesystem_browser import BaseFilesystemBrowser


@override_settings(DEBUG=False)
class TestFilesystemBrowser(SimpleTestCase):
    """
    e.g.:
    https://en.wikipedia.org/wiki/Directory_traversal_attack
    """
    BASE_PATH=os.path.abspath(os.path.dirname(django_tools.__file__))

    def test_directory_traversal_attack1(self):
        try:
            BaseFilesystemBrowser(request=None, absolute_path=self.BASE_PATH, base_url="bar", rest_url="../")
        except Http404 as err:
            self.assertEqual(six.text_type(err), "Directory doesn't exist!")

    @override_settings(DEBUG=True)
    def test_debug_message(self):
        sub_path = os.path.normpath(os.path.join(self.BASE_PATH, ".."))
        try:
            BaseFilesystemBrowser(request=None, absolute_path=self.BASE_PATH, base_url="bar", rest_url="../")
        except Http404 as err:
            self.assertEqual(
                err.args[0].message,
                "Directory '%s' is not in base path ('%s')" % (sub_path, self.BASE_PATH)
            )

    def test_directory_traversal_attack_encodings(self):
        base_path=os.path.abspath(os.path.dirname(django_tools.__file__))

        rest_urls = (
            "/etc/passwd",
            "..",
            "../",
            "\\\\",

            # URI encoded directory traversal:
            "%2e%2e%2f", # ../
            "%2e%2e/", # ../
            "..%2f", # ../
            "%2e%2e%5c", # ..\

            # Unicode / UTF-8 encoded directory traversal:
            "..%c1%1c", # ../
            "..%c0%af", # ..\
            "%c0%ae%c0%ae%c1%1c", # %c0%ae -> . -> ../
            "%c0%ae%c0%ae%c0%af", # %c0%ae -> . -> ..\
        )
        for rest_url in rest_urls:
            # print(rest_url)
            try:
                BaseFilesystemBrowser(request=None, absolute_path=self.BASE_PATH, base_url="bar", rest_url=rest_url)
            except Http404 as err:
                self.assertEqual(six.text_type(err), "Directory doesn't exist!")



"""
    :created: 2017 by Jens Diemer
    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import unittest

from django.http.request import validate_host, split_domain_port, host_validation_re

# https://github.com/jedie/django-tools
from django_tools.settings_utils import FnMatchIps


class TestSettingsUtils(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.fnmatch_ips = FnMatchIps(["127.0.0.1", "::1", "192.168.*.*"])

    def test_contains(self):
        # INTERNAL_IPS used this
        self.assertTrue(
            "192.168.1.2" in self.fnmatch_ips
        )
        self.assertFalse(
            "10.0.1.2" in self.fnmatch_ips
        )

    def test_compare(self):
        # ALLOWED_HOSTS with django.http.request.validate_host

        self.assertTrue(
            validate_host("192.168.1.2", self.fnmatch_ips)
        )

        self.assertFalse(
            validate_host("10.0.1.2", self.fnmatch_ips)
        )

    def test_str(self):
        self.assertEqual(
            "127.0.0.1", str(self.fnmatch_ips[0])
        )

    def test_re_usage(self):
        """
        e.g.: django.http.request.split_domain_port used RE in test environment
        :return:
        """
        host = self.fnmatch_ips[0]

        self.assertTrue(host_validation_re.match(host))

        domain, port = split_domain_port(host)
        self.assertEqual(domain, "127.0.0.1")
        self.assertEqual(port, "")




import logging
import sys

from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import UserCreationForm
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.http import HttpRequest

log = logging.getLogger(__name__)


def create_user(
    username,
    password=None,
    email="",
    is_staff=False,
    is_superuser=False,
    encrypted_password=None,
    groups=None,
    update_existing=False,
):
    """
    Create a user and return the instance.

    :param password: Plain Text password
    :param encrypted_password: Hashed password (e.g. cloned vom other user)
    :param groups: List of user group names
    :return: created user instance
    """
    if password is None and encrypted_password is None:
        raise RuntimeError("'password' or 'encrypted_password' needed.")

    User = get_user_model()

    user = None
    if update_existing:
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            user = None
        else:
            created = False

    if user is None:
        # Create user with origin form to validate username:
        user_form_data = {"username": username, "password1": password, "password2": password}

        if password is None:
            # encrypted_password set later!
            # Set something, to run validation.
            user_form_data["password1"] = user_form_data["password2"] = "A temp password!"

        user_form = UserCreationForm(user_form_data)
        if user_form.is_valid():
            user = user_form.save(commit=False)
            created = True
        else:
            raise ValidationError("%s" % user_form.errors)

    user.email = email
    user.is_staff = is_staff
    user.is_superuser = is_superuser

    if encrypted_password:
        log.critical("Set encrypted user password, ok.")
        user.password = encrypted_password
    elif password:
        user.set_password(password)
    else:
        raise RuntimeError  # should never happen, see above ;)

    user.save()

    if groups:
        for group_name in groups:
            try:
                user_group = Group.objects.get(name=group_name)
            except Group.DoesNotExist:
                raise RuntimeError('User group "%s" not found!' % group_name)
            user.groups.add(user_group)

    if created:
        log.debug("Create new user %r account" % user)
    else:
        log.debug("Update existing %r account" % user)

    return user


def get_super_user():
    """
    :return: the first 'superuser'
    """
    User = get_user_model()
    try:
        super_user = User.objects.filter(is_superuser=True)[0]
    except (User.DoesNotExist, IndexError):
        log.error("Can't get a superuser. Please create first!")
        raise RuntimeError("No superuser")
    return super_user


def get_or_create_user(username, group, encrypted_password):
    User = get_user_model()
    try:
        user = User.objects.get(username=username)
    except User.DoesNotExist:
        user = create_user(
            username=username,
            groups=(group,),
            is_staff=True,
            is_superuser=False,
            encrypted_password=encrypted_password,
        )
        created = True
    else:
        created = False

    user.groups = (group,)
    user.is_staff = True
    user.is_superuser = False
    user.save()

    return user, created


def get_or_create_group(groupname, permissions):
    group, created = Group.objects.get_or_create(name=groupname)

    for permission in group.permissions.all():
        if permission not in permissions:
            print("remove permission:", permission)
            group.permissions.remove(permission)

    print("Add %i permissions to %r" % (len(permissions), groupname))
    for permission in permissions:
        group.permissions.add(permission)

    existing_permission_count = group.permissions.all().count()
    print("Group %s has %i permissions" % (groupname, existing_permission_count))
    assert len(permissions) == existing_permission_count, "Wrong permission count: %i != %i" % (
        existing_permission_count,
        len(permissions),
    )

    return group, created


def get_or_create_user_and_group(username, groupname, permissions, encrypted_password):
    """
    * Creates a user group if not exists
    * add given permissions to user group
    * create user if not exists
    * adds user into user group

    e.g.:
        from django_tools.permissions import get_filtered_permissions
        from django_tools.unittest_utils.user import get_or_create_user_and_group

        superuser = User.objects.filter(is_superuser=True, is_active=True)[0]
        encrypted_password = superuser.password

        test_user = get_or_create_user_and_group(
            username="testuser",
            groupname="testgroup",
            permissions=get_filtered_permissions(
                exclude_app_labels=("auth", "sites"),
                exclude_models=(),
                exclude_permissions=(),
            ),
            encrypted_password=encrypted_password
        )
    """
    log.info("Create user group %s and add %i permissions...", groupname, len(permissions))
    group, created = get_or_create_group(groupname, permissions)
    if created:
        log.debug("User group '%s' created.", groupname)
    else:
        log.debug("Use existing user group '%s', ok.", groupname)

    log.info("Create user '%s'...", username)
    user, created = get_or_create_user(username, group, encrypted_password)
    if created:
        log.debug("User '%s' created.", username)
    else:
        log.debug("User '%s' already exists, ok.", username)

    return user


class TestUserMixin:
    """
    Important: Test user will only be created, if django TestCase is used!
    e.g:

    from django.test.testcases import TestCase

    class FooBarTestCase(TestUserMixin, TestCase):
        def test...

    """

    TEST_USERS = {
        "superuser": {
            "username": "superuser",
            "email": "superuser@example.org",
            "password": "superuser_password",
            "is_staff": True,
            "is_superuser": True,
        },
        "staff": {
            "username": "staff_test_user",
            "email": "staff_test_user@example.org",
            "password": "staff_test_user_password",
            "is_staff": True,
            "is_superuser": False,
        },
        "normal": {
            "username": "normal_test_user",
            "email": "normal_test_user@example.org",
            "password": "normal_test_user_password",
            "is_staff": False,
            "is_superuser": False,
        },
    }

    @classmethod
    def setUpClass(cls):
        super(TestUserMixin, cls).setUpClass()
        cls.UserModel = auth.get_user_model()

    @classmethod
    def setUpTestData(cls):
        super(TestUserMixin, cls).setUpTestData()
        cls.create_testusers(cls)

    def create_testusers(self):
        """
        Create all available testusers and UserProfiles
        """
        for user_data in self.TEST_USERS.values():
            user = create_user(update_existing=True, **user_data)
            log.debug("Test user %s created." % user)

    def get_userdata(self, usertype):
        """ return userdata from self.TEST_USERS for the given usertype """
        try:
            return self.TEST_USERS[usertype]
        except KeyError as err:
            etype, evalue, etb = sys.exc_info()
            evalue = etype(
                "Wrong usetype %s! Existing usertypes are: %s" % (err, ", ".join(list(self.TEST_USERS.keys())))
            )
            raise etype(evalue).with_traceback(etb)

    def _get_user(self, usertype):
        """ return User model instance for the given usertype"""
        test_user = self.get_userdata(usertype)
        return self.UserModel.objects.get(username=test_user["username"])

    def login(self, usertype):
        """
        Login the user defined in self.TEST_USERS
        return User model instance
        """
        test_user = self.get_userdata(usertype)

        username = test_user["username"]

        count = self.UserModel.objects.all().count()
        self.assertNotEqual(count, 0, "You have to call self.create_testusers() first!")

        # Some auth backends needs request object (e.g.: django-axes)
        request = HttpRequest()
        user = self.UserModel.objects.get(username=username)
        request.user = user

        ok = self.client.login(request=request, username=test_user["username"], password=test_user["password"])
        self.assertTrue(ok, 'Can\'t login test user "%s"!' % usertype)

        return self._get_user(usertype)

    def add_user_permissions(self, user, permissions):
        """
        add permissions to the given user instance.
        permissions e.g.: ("AppLabel.add_Modelname", "auth.change_user")
        """
        assert isinstance(permissions, (list, tuple))
        for permission in permissions:
            # permission, e.g: blog.add_blogentry
            app_label, permission_codename = permission.split(".", 1)
            model_name = permission_codename.split("_", 1)[1]

            try:
                content_type = ContentType.objects.get(app_label=app_label, model=model_name)
            except ContentType.DoesNotExist:
                etype, evalue, etb = sys.exc_info()
                evalue = etype(
                    'Can\'t get ContentType for app "%s" and model "%s": %s' % (app_label, model_name, evalue)
                )
                raise etype(evalue).with_traceback(etb)

            perm = Permission.objects.get(content_type=content_type, codename=permission_codename)
            user.user_permissions.add(perm)
            user.save()

        self.assertTrue(user.has_perms(permissions))

    def refresh_user(self, user):
        """
        Return a fresh User model instance from DB.
        Note: Using "user.refresh_from_db()" will not help in every case!
        e.g.: Add new permission on user group and check the added one.
        """
        return self.UserModel.objects.get(pk=user.pk)



# coding: utf-8

"""
    unittests for filemanager
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2012-2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import os
import unittest
import tempfile

from django.http import Http404

from django_tools.filemanager.filemanager import BaseFilemanager
from django_tools.filemanager.exceptions import DirectoryTraversalAttack


class FilemanagerBaseTestCase(unittest.TestCase):
    """
    Create this test filesystem tree:

    self.BASE_PATH/file1.txt
    self.BASE_PATH/file2.txt
    self.BASE_PATH/subdir1/subfile1.txt
    self.BASE_PATH/subdir1/subsubdir1/
    self.BASE_PATH/subdir1/subsubdir1/subsubfile1.txt
    self.BASE_PATH/subdir1/subsubdir1/subsubfile2.txt
    self.BASE_PATH/subdir1/emptysubsubdir/
    self.BASE_PATH/emptysubdir/
    """
    @classmethod
    def setUpClass(cls):
        super(FilemanagerBaseTestCase, cls).setUpClass()
        cls.BASE_PATH = "%s/" % tempfile.mkdtemp(prefix="filemanager-unittests_")

        # For dir traversal attack tests
        cls.SUB_BASE_PATH = os.path.join(cls.BASE_PATH, "subdir1")
        cls.ATTACK_PARTS = ("..", "..\\", "../",
            # from https://en.wikipedia.org/wiki/Directory_traversal_attack#URI_encoded_directory_traversal
            "%2e%2e%2f", "%2e%2e/", "..%2f", "%2e%2e%5c"
            # from https://en.wikipedia.org/wiki/Directory_traversal_attack#Unicode_.2F_UTF-8_encoded_directory_traversal
            "..%c1%1c", "..%c0%af",
            "..\xc1\x1c", "..\xc0\xaf",
        )

        cls.DIRS = (
            (cls.BASE_PATH, "subdir1", "subsubdir1"),
            (cls.BASE_PATH, "subdir1", "emptysubsubdir"),
            (cls.BASE_PATH, "emptysubdir"),
        )
        cls.FILES = (
            (cls.BASE_PATH, "file1.txt"),
            (cls.BASE_PATH, "file2.txt"),
            (cls.BASE_PATH, "subdir1", "subfile1.txt"),
            (cls.BASE_PATH, "subdir1", "subsubdir1", "subsubfile1.txt"),
            (cls.BASE_PATH, "subdir1", "subsubdir1", "subsubfile2.txt"),
        )

        # Create dirs:
        for dirs in cls.DIRS:
            path = os.path.join(*dirs)
#            print "Create dir: %s" % repr(path)
            os.makedirs(path)

        # Create files:
        for file_info in cls.FILES:
            path = os.path.join(*file_info)
            content = "File content for: %s" % repr(file_info[-1])
#            print "Create file %s with %s" % (repr(path), repr(content))
            with open(path, "w") as f:
                f.write(content)

    @classmethod
    def tearDownClass(cls):
        super(FilemanagerBaseTestCase, cls).tearDownClass()
        tempdir = tempfile.gettempdir()
        if cls.BASE_PATH.startswith(tempdir):
            for root, dirs, files in os.walk(cls.BASE_PATH, topdown=False):
                for name in files:
                    path = os.path.join(root, name)
                    os.remove(path)
                for name in dirs:
                    path = os.path.join(root, name)
                    os.rmdir(path)
            os.rmdir(cls.BASE_PATH)
        else:
            cls.fail("Cleanup error: %s not in %s" % (cls.BASE_PATH, tempdir))

    def test_setup(self):
        """ if this tests fails, all other may fail, too. """
        for file_info in self.FILES:
            path = os.path.join(*file_info)
            self.assertTrue(os.path.isfile(path))


class FilemanagerDirectoryTraversal(FilemanagerBaseTestCase):

    def test_subdir1(self):
        subdir = "subdir1"
        base_url = "/base/url/"
        rest_url = "%s/" % subdir
        fm = BaseFilemanager(None, self.BASE_PATH, base_url, rest_url)
        self.assertEqual(fm.abs_url, '/base/url/subdir1/')
        self.assertEqual(fm.absolute_path, self.BASE_PATH)
        self.assertEqual(fm.breadcrumbs, [
            {'url': base_url, 'name': 'index', 'title': "goto 'index'"},
            {'url': '/base/url/%s/' % subdir, 'name': subdir, 'title': "goto '%s'" % subdir}
        ])

    def test_not_existing_path1(self):
        self.assertRaises(Http404,
            BaseFilemanager, None, self.BASE_PATH, "/base/url/", "not_exists"
        )
    def test_not_existing_path2(self):
        self.assertRaises(Http404,
            BaseFilemanager, None, self.BASE_PATH, "/base/url/", "does/not/exists"
        )

    def test_dir_traversal_attack1(self):
        for parts in self.ATTACK_PARTS:
            self.assertRaises((DirectoryTraversalAttack, Http404),
                BaseFilemanager, None, self.SUB_BASE_PATH, "/base/url/", parts,
            )

    def test_dir_traversal_attack2(self):
        for parts in self.ATTACK_PARTS:
            self.assertRaises((DirectoryTraversalAttack, Http404),
                BaseFilemanager, None, self.SUB_BASE_PATH, "/base/url/", "%semptysubdir" % parts,
            )

    def test_dir_traversal_attack3(self):
        self.assertRaises(Http404,
            BaseFilemanager, None, self.BASE_PATH, "/base/url/", "subdir1/../../etc/passwd"
        )

    def test_dir_traversal_attack4(self):
        self.assertRaises(Http404,
            BaseFilemanager, None, self.BASE_PATH, "/base/url/", "subdir1/c:\\boot.ini"
        )





# coding:utf-8

from importlib import import_module

from django.utils.six import PY2
from django.conf import settings
from django.core import urlresolvers



def get_filtered_apps(resolve_url="/", no_args=True, debug=False, skip_fail=False):
    """
    Filter settings.INSTALLED_APPS and create a list
    of all Apps witch can resolve the given url >resolve_url<

    @param resolve_url: url used for RegexURLResolver
    @param no_args: Only views without args/kwargs ?
    @parm skip_fail: If True: raise exception if app is not importable

    Please look at:

        django_tools.django_tools_tests.test_installed_apps_utils
    
    with debug, some print messages would be created:

    e.g.: get_filtered_apps(debug=True)
    found 'django.contrib.admindocs' with urls.py
    found 'django.contrib.auth' with urls.py
    Skip 'django.contrib.auth': Can't handle root url.
    found 'django.contrib.flatpages' with urls.py
    ['django.contrib.admindocs']
    """
    root_apps = []
    for app_label in settings.INSTALLED_APPS:
        urls_pkg = "%s.urls" % app_label
        try:
            url_mod = import_module(urls_pkg)
        except ImportError as err:
            if debug:
                print("Skip %r: has no urls.py" % app_label)
            if PY2:
                msg_should = "No module named urls"
            else:
                msg_should = "No module named '%s'" % urls_pkg
            if str(err) == msg_should:
                continue
            if not skip_fail:
                raise
        except Exception as err:
            if debug:
                print("Error importing %r: %s" % (app_label, err))
            if not skip_fail:
                raise
            else:
                continue

        if debug:
            print("found %r with urls.py" % app_label)

        try:
            urlpatterns = url_mod.urlpatterns
        except AttributeError:
            if debug:
                print("Skip %r: urls.py has no 'urlpatterns'" % app_label)
            continue

        resolver = urlresolvers.RegexURLResolver(r'^', urlpatterns)
        try:
            func, func_args, func_kwargs = resolver.resolve(resolve_url)
        except urlresolvers.Resolver404 as err:
            if debug:
                print("Skip %r: Can't handle root url." % app_label)
            continue
        if not no_args or func_args == () and func_kwargs == {}:
            root_apps.append(app_label)
    return root_apps


if __name__ == "__main__":
    import doctest
    doctest.testmod(
#        verbose=True
        verbose=False
    )
    print("DocTest end.")



# coding: utf-8

"""
    :copyleft: 2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


from django.test import SimpleTestCase

from django_tools.template.render import render_string_template, \
    render_template_file


class TestRender(SimpleTestCase):
    def test_render_template_file(self):
        context={"foo":"bar"}
        path = "test_template.html"
        x = render_template_file(path, context)
        print(x)

        # Note: START/END comments added by: django_tools.template.loader.DebugCacheLoader

        self.assertEqual(x, (
            "<!-- START 'test_template.html' -->\n"
            "Hello bar !\n\n"
            "<!-- END 'test_template.html' -->"
        ))

    def test_render_string_template(self):
        x = render_string_template("Foo {{ bar }}!", {"bar": "BAR"})
        self.assertEqual(x, 'Foo BAR!')



# coding: utf-8

"""
    media path selection
    ~~~~~~~~~~~~~~~~~~~~

    TODO: Made this generic and don't use settings.MEDIA_ROOT direct, let it
    be set as a argument to widget/fields etc.

     * model field
     * form field
     * widget

     INFO: This exist only for backward-compatibility and will be removed
     in the future. Please use static_path!

    :copyleft: 2010-2016 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import os
import warnings

from django import forms
from django.db import models
from django.conf import settings
from django.utils.six import with_metaclass

from django_tools.utils.messages import failsafe_message


def directory_walk(path):
    """
    Directory tree generator
    similar to os.walk, except:
      - yield only directories
      - sorted list case-insensitive
      - follow links (os.walk can do this since python 2.6)
    """
    dirs = []
    for name in os.listdir(path):
        if os.path.isdir(os.path.join(path, name)):
            dirs.append(name)

    yield path

    # Sort case-insensitive
    dirs.sort(key=str.lower)

    for dir in dirs:
        sub_path = os.path.join(path, dir)
        for x in directory_walk(sub_path):
            yield x




class MediaPathWidget(forms.Select):
    """
    Select a sub directory in settings.MEDIA_ROOT

    >>> import os, django_tools
    >>> settings.MEDIA_ROOT = os.path.dirname(os.path.abspath(django_tools.__file__))
    >>> MediaPathWidget().choices[:2]
    [('__pycache__', '__pycache__'), ('admin_tools', 'admin_tools')]
    """
    def __init__(self, attrs=None):
        super(MediaPathWidget, self).__init__(attrs)

        self._base_path = os.path.abspath(os.path.normpath(settings.MEDIA_ROOT))

        try:
            self.choices = self._get_path_choices()
        except OSError as err:
            self.choices = []
            if settings.DEBUG:
                failsafe_message("Can't read MEDIA_ROOT: %s" % err)

        warnings.warn(
            "MediaPathWidget is deprecated and will removed in the future!"
            " Please use StaticPathWidget.",
            PendingDeprecationWarning
        )

    def _get_path_choices(self):
        media_dirs_choices = []
        cut_pos = len(self._base_path)
        for root in directory_walk(self._base_path):
            rel_dir = root[cut_pos:].strip(os.sep)
            if rel_dir:
                media_dirs_choices.append((rel_dir, rel_dir))
        return media_dirs_choices


class MediaPathModelField(models.TextField):#, with_metaclass(models.SubfieldBase)):
    """

    """

#    def __init__(self, separator=",", strip_items=True, skip_empty=True, *args, **kwargs):
#        self.separator = separator
#        self.strip_items = strip_items
#        self.skip_empty = skip_empty
#        super(MediaPathModelField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        """ Use always own widget and form field. """

        warnings.warn(
            "MediaPathModelField is deprecated and will removed in the future!"
            " Please use StaticPathModelField.",
            PendingDeprecationWarning
        )

        kwargs["widget"] = MediaPathWidget
#        kwargs["form_class"] = SignSeparatedFormField
        return super(MediaPathModelField, self).formfield(**kwargs)





# coding: utf-8

"""
    utils for filemanager
    ~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function



import posixpath
import os
import ntpath


def clean_posixpath(path, up_level_references=False):
    """
    Based on code from:
    https://github.com/django/django/blob/master/django/views/static.py
    
    But we build always a posixpath with "/" as path separate character.  
    
    
    Keep starting/ending slash:
    
    >>> clean_posixpath("no/slash")
    'no/slash'
    >>> clean_posixpath("/starts/with/slash")
    '/starts/with/slash'
    >>> clean_posixpath("ends/with/slash/")
    'ends/with/slash/'
    >>> clean_posixpath("") # normpath would return: "."
    ''
    >>> clean_posixpath("/")
    '/'
    >>> clean_posixpath("/../")
    '/'


    Remove every crude characters:
    
    >>> clean_posixpath("foo//bar")
    'foo/bar'
    >>> clean_posixpath("/foo/./bar")
    '/foo/bar'
    >>> clean_posixpath(r"foo\\bar/")
    'foo/bar/'
    
    
    up-level references would be only applied if activated:
    
    >>> clean_posixpath("/foo/bar/../../etc/passwd") # normpath would return: '../etc/passwd'
    '/foo/bar/etc/passwd'
    >>> clean_posixpath("/foo/bar/../../etc/passwd", up_level_references=True)
    '/etc/passwd'
    
    >>> clean_posixpath("../../../etc/passwd") # normpath would return: '../../../etc/passwd'
    'etc/passwd'
    
    >>> clean_posixpath(r"\\foo\\bar\\..\\etc\\passwd") # normpath would return: '\\foo\\bar\\..\\etc\\password'
    '/foo/bar/etc/passwd'
    
    
    Ignore windows drive parts:
    
    >>> clean_posixpath(r"c:\\boot.ini")
    'boot.ini'
    >>> clean_posixpath(r"foo/bar/c:\\boot.ini")
    'foo/bar/boot.ini'
    """
    path = path.replace('\\', '/')

    add_slash = path.endswith("/")    
    if path.startswith("/"):
        newpath = "/"
    else:
        newpath = ""    
    path = path.strip("/")

    if up_level_references:
        # e.g.: foo/../bar -> bar
        path = posixpath.normpath(path)

    for part in path.split("/"):       
        if not part:
            # Strip empty path components.
            continue

        drive, part = ntpath.splitdrive(part)
        head, part = posixpath.split(part)
        if part in (".", ".."):
            continue
        newpath = posixpath.join(newpath, part)
        
    if add_slash and newpath != "/":
        newpath += "/"
        
    return newpath


def add_slash(path):
    """
    >>> add_slash("/foo")
    '/foo/'
    >>> add_slash("/bar/")
    '/bar/'
    """
    if not path.endswith(os.sep):
        path += os.sep
    return path    


def symbolic_notation(mode):
    """
    Convert os.stat().st_mode values to a symbolic representation. e.g: 
       
    >>> s = symbolic_notation(16893) # -> 040775 -> 775
    >>> s == 'rwxrwxr-x'
    True
    
    >>> s = symbolic_notation(33204) # -> 0100664 -> 664
    >>> s == 'rw-rw-r--'
    True
    """
    mode = mode & 0o777 # strip "meta info"
    chmod_symbol = ''.join(
        mode & 0o400 >> i and x or '-' for i, x in enumerate('rwxrwxrwx')
    )
    return chmod_symbol


if __name__ == "__main__":
    import doctest
    print(doctest.testmod(
#        verbose=True
        verbose=False
    ))


# coding: utf-8

"""
    forms utils
    ~~~~~~~~~~~

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author:$

    :copyleft: 2009 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


from django import forms


class LimitManyToManyFields(object):
    """
    Limit ManyToMany fields in forms. Hide the field, if only one item can be selected.
    
    e.g. limit sites choices only to accessible sites:
    --------------------------------------------------------------------------
    class MyModel(models.Model):
        sites = models.ManyToManyField(Site)
        ...
        
    class UserProfile(models.Model):
        sites = models.ManyToManyField(Site)
        ...
        
    class MyForm(LimitManyToManyFields, forms.ModelForm): # <- Order is important !
        class Meta:
            model = MyModel
            
    def my_view(request):
        user_profile = request.user.get_profile()

        m2m_limit = {"sites": user_profile.sites.values_list("id", "name")}
        
        if request.method == "POST":
            form = MyForm(m2m_limit, request.POST)
            if form.is_valid():
                ...
        else:
            # Preselect all existing sites
            form = MyForm(m2m_limit)
        ...
    --------------------------------------------------------------------------
    
    crosspost: http://www.djangosnippets.org/snippets/1692/
    """
    def __init__(self, m2m_limit, *args, **kwargs):
        """
        preselect site select options. If user can only access one site or there exist only one site
        we remove the site field and insert the site info in save() method.
        """
        assert isinstance(m2m_limit, dict), \
            "%s error: first argument must be the m2m limit dict!" % self.__class__.__name__

        super(LimitManyToManyFields, self).__init__(*args, **kwargs)

        for field_name, limits in m2m_limit.items():
            if len(limits) == 1:
                value = int(limits[0][0])
                # Only one item can be selected. Hide the ManyToMany field. To hide the field and
                # for validation, we changed the MultipleChoiceField to a IntegerField.
                self.fields[field_name] = forms.IntegerField(
                    max_value=value, min_value=value, initial=value
                )
#                self.fields[field_name].widget.input_type = 'hidden'
#                self.fields[field_name].widget.is_hidden = True
            else:
                # Limit the ManyToMany field choices
                self.fields[field_name].choices = limits





"""
    TODO: remove this file in future!

    :created: 2015 by Jens Diemer
    :copyleft: 2015-2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import logging

# https://github.com/jedie/django-tools
from django_tools.decorators import warn_class_usage, warn_function_usage
from django_tools.selenium.chromedriver import chromium_available as new_chromium_available
from django_tools.selenium.django import (
    SeleniumChromiumStaticLiveServerTestCase, SeleniumFirefoxStaticLiveServerTestCase
)
from django_tools.selenium.geckodriver import firefox_available as new_firefox_available
from django_tools.selenium.utils import find_executable as new_find_executable

log = logging.getLogger(__name__)


@warn_class_usage("Use 'from django_tools.selenium.django import SeleniumChromiumStaticLiveServerTestCase' !")
class SeleniumChromiumTestCase(SeleniumChromiumStaticLiveServerTestCase):
    pass


@warn_class_usage("Use 'from django_tools.selenium.django import SeleniumFirefoxStaticLiveServerTestCase' !")
class SeleniumFirefoxTestCase(SeleniumFirefoxStaticLiveServerTestCase):
    pass


@warn_function_usage("Use 'from django_tools.selenium.utils import find_executable' !")
def find_executable(*args, **kwargs):
    return new_find_executable(*args, **kwargs)


@warn_function_usage("Use 'from django_tools.selenium.geckodriver import firefox_available' !")
def firefox_available(*args, **kwargs):
    return new_firefox_available(*args, **kwargs)


@warn_function_usage("Use 'from django_tools.selenium.chromedriver import chromium_available' !")
def chromium_available(*args, **kwargs):
    return new_chromium_available(*args, **kwargs)



# coding:utf-8

"""
    SlowerDevServer
    ~~~~~~~~~~~~~~~
    
    Simple slow down the django developer server.
    The middleware insert in every 200 response a time.sleep
    
    Put this into your settings:   
    --------------------------------------------------------------------------
        MIDDLEWARE_CLASSES = (
            ...
            'django_tools.middlewares.SlowerDevServer.SlowerDevServerMiddleware',
            ...
        )
        SLOWER_DEV_SERVER_SLEEP = 0.3 # time.sleep() value (in sec.)
    --------------------------------------------------------------------------
    
    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate$
    $Rev$
    $Author:$

    :copyleft: 2009 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import time
import warnings

from django.conf import settings

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object  # fallback for Django < 1.10


class SlowerDevServerMiddleware(MiddlewareMixin):
    def __init__(self):
        warnings.warn("Slower developer server used!")
        
    def process_response(self, request, response):
        if response.status_code == 200: 
            print("SlowerDevServerMiddleware: Wait for %sSec..." % settings.SLOWER_DEV_SERVER_SLEEP)
            time.sleep(settings.SLOWER_DEV_SERVER_SLEEP)
        return response



import os
import shutil
import tempfile
from pathlib import Path

from django.test.utils import TestContextDecorator


class isolated_filesystem(TestContextDecorator):
    """
    Acts as either a decorator or a context manager.

    with isolated_filesystem(prefix="temp_dir_prefix"):
        print("I'm in the temp path here: %s" % Path().cwd())

    or:

    class FooBarTestCase(TestCase):
        @isolated_filesystem()
        def test_foo(self):
            print("I'm in the temp path here: %s" % Path().cwd())

    or:

    @isolated_filesystem()
    class FooBarTestCase(TestCase):
        def test_foo(self):
            print("I'm in the temp path here: %s" % Path().cwd())
    """
    def __init__(self, prefix=None):
        super().__init__()

        self.prefix = prefix
        self.prefix_candidate = None

    def enable(self):
        if self.prefix is None:
            prefix = self.prefix_candidate
        else:
            prefix = self.prefix

        print("Use prefix: %r" % prefix)

        self.cwd = Path().cwd()
        self.temp_path = tempfile.mkdtemp(prefix=prefix)
        os.chdir(self.temp_path)

    def disable(self):
        os.chdir(str(self.cwd)) # str() needed for older python <=3.5
        try:
            shutil.rmtree(self.temp_path)
        except (OSError, IOError):
            pass

    def decorate_class(self, cls):
        self.prefix_candidate = cls.__name__
        print("prefix from class %r: %r" % (cls, self.prefix_candidate))
        return super().decorate_class(cls)

    def decorate_callable(self, func):
        self.prefix_candidate = func.__name__
        print("prefix from func %r: %r" % (func, self.prefix_candidate))
        return super().decorate_callable(func)



# coding: utf-8

"""
    Test django_tools.utils.url
"""

from __future__ import unicode_literals

from django.test import SimpleTestCase
from django_tools.utils.url import GetDict


class TestGetDict(SimpleTestCase):
    def test_django_example(self):
        q = GetDict()
        q['next'] = '/a&b/'
        self.assertEqual(q.urlencode(), 'next=%2Fa%26b%2F')
        self.assertEqual(q.urlencode(safe='/'), 'next=/a%26b/')

    def test_empty(self):
        q = GetDict()
        q['this_is_empty'] = None
        self.assertEqual(q.urlencode(), 'this_is_empty')
        q['another_empty'] = None
        self.assertEqual(q.urlencode(), 'another_empty&this_is_empty')

    def test_multi(self):
        q = GetDict(str('vote=yes&vote=no'))
        self.assertEqual(q.urlencode(), 'vote=no&vote=yes')
        q['empty'] = None
        self.assertEqual(q.urlencode(), 'empty&vote=no&vote=yes')

    def test_enhanced_example(self):
        q = GetDict()
        q['next'] = '/a&b/'
        q['foo'] = 'bar'
        q['number'] = 123
        q['empty'] = None
        self.assertEqual(q.urlencode(), 'empty&foo=bar&next=%2Fa%26b%2F&number=123')
        self.assertEqual(q.urlencode(safe='/'), 'empty&foo=bar&next=/a%26b/&number=123')



# coding: utf-8

"""
    filemanager
    ~~~~~~~~~~~

    :copyleft: 2012-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

import os
import posixpath

from django.utils.six.moves import urllib
from django.conf import settings
from django.core.exceptions import ValidationError
from django.http import Http404
from django.utils.translation import ugettext as _

from django_tools.filemanager.utils import add_slash, clean_posixpath
from django_tools.filemanager.exceptions import DirectoryTraversalAttack
from django_tools.validators import ExistingDirValidator

STOP_PARTS = (
    # https://en.wikipedia.org/wiki/Directory_traversal_attack#Unicode_.2F_UTF-8_encoded_directory_traversal
    "%c1%1c", # %c1%1c can be translated to /
    "%c0%af", # %c0%af can be translated to \
    "%c0%ae", # %c0%af can be translated to .
)

class BaseFilesystemBrowser(object):
    """
    Base class for a django app like a filemanager, which contains only
    the base functionality to browse to a base path of the filesystem.
    """
    def __init__(self, request, absolute_path, base_url, rest_url):
        """
        absolute_path - path in filesystem to the root directory
        base_url - url prefix of this filemanager instance
        rest_url - relative sub path of the current view

        it is assumed that 'absolute_path' and 'base_url' are internal values
        and 'rest_url' are a external given value from the requested user.

        TODO: Use django_tools.validators.ExistingDirValidator and merge code!
        """
        self.request = request
        self.absolute_path = add_slash(absolute_path)
        self.base_url = clean_posixpath(base_url)

        self.dir_validator = ExistingDirValidator(self.absolute_path)

        rest_url = add_slash(rest_url)
        try:
            self.dir_validator(rest_url)
        except ValidationError as err:
            if settings.DEBUG:
                raise Http404(err)
            else:
                raise Http404(_("Directory doesn't exist!"))

        self.rel_url = posixpath.normpath(rest_url).lstrip("/")
        self.abs_url = posixpath.join(self.base_url, rest_url)
        if not os.path.isdir(self.absolute_path):
            if settings.DEBUG:
                raise Http404(
                    "Formed path %r doesn't exist." % self.absolute_path)
            else:
                raise Http404(_("Directory doesn't exist!"))

        # # print("rest_url 1: %r" % rest_url)
        # for part in STOP_PARTS:
        #     if part in rest_url:
        #         raise DirectoryTraversalAttack("Stop chars %r found!" % part)
        #
        # rest_url = urllib.parse.unquote(rest_url)
        # # print("rest_url 2: %r" % rest_url)
        #
        #
        #
        # # To protect from directory traversal attack
        # # https://en.wikipedia.org/wiki/Directory_traversal_attack
        # clean_rest_url = clean_posixpath(rest_url)
        # if clean_rest_url != rest_url:
        #     # path changed cause of "illegal" characters
        #     raise DirectoryTraversalAttack(
        #         "path %s is not equal to cleaned path: %s" % (repr(rest_url), repr(clean_rest_url))
        #     )
        #
        # self.rel_url = rest_url.lstrip("/")
        # self.rel_path = add_slash(os.path.normpath(self.rel_url))
        #
        # self.abs_path = clean_posixpath(os.path.join(self.absolute_path, self.rel_path))
        # self.check_path(self.absolute_path, self.abs_path)
        #
        # self.abs_url = posixpath.join(self.base_url, self.rel_url)
        #
        # if not os.path.isdir(self.abs_path):
        #     raise Http404("Formed path %r doesn't exist." % self.abs_path)

        self.breadcrumbs = self.build_breadcrumbs()

    def build_breadcrumbs(self):
        parts = ""
        url = self.base_url
        breadcrumbs = [{
            "name": _("index"),
            "title": _("goto 'index'"),
            "url": url
        }]
        rel_url = self.rel_url.strip("/")
        if not rel_url:
            return breadcrumbs

        for url_part in rel_url.split("/"):
            url += "%s/" % url_part
            parts += "%s/" % url_part
            breadcrumbs.append({
                "name": url_part,
                "title": _("goto '%s'") % parts.strip("/"),
                "url": url
            })
        return breadcrumbs

    def check_path(self, base_path, path):
        """
        Simple check if the path is a sub directory of base_path.
        This must be called from external!
        """
        # Important: the path must be terminated with a slash, otherwise:
        #
        # base_path = /foo/bar
        # path      = /foo/barNEW
        #
        # path starts with base_path without slashes
        #
        # which add slash:
        #
        # base_path = /foo/bar    -> /foo/bar/
        # path      = /foo/barNEW -> /foo/barNEW/
        #              doesn't start with ---^
        #
        assert base_path.endswith(os.sep), "'base_path' must ended with a slash!"
        assert path.endswith(os.sep), "'path' must ended with a slash!"

        if not path.startswith(base_path):
            raise DirectoryTraversalAttack("%r doesn't start with %r" % (path, base_path))





# coding:utf-8

"""
    utils around django messages
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    http://docs.djangoproject.com/en/dev/ref/contrib/messages/
    
    :copyleft: 2010-2011 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


import inspect
import warnings

from django.conf import settings
from django.contrib import messages
from django.contrib.messages.storage.fallback import FallbackStorage

from django_tools.middlewares import ThreadLocal
from django_tools.utils.stack_info import get_stack_info


STACK_LIMIT = 6 # Display only the last X stack lines
MAX_FILEPATH_LEN = 50 # Cut filepath in stack info message


class FileLikeMessages(object):
    """
    Simple layer around messages, to get a file-like object.
    
    usage e.g.:
    
    page_msg = FileLikeMessages(request, messages.INFO)
    page_msg.write("This is it!")
    page_msg("Call works's, too.")
    """
    def __init__(self, request, msg_level):
        self.request = request
        self.msg_level = msg_level

    def write(self, txt):
        messages.add_message(self.request, self.msg_level, txt)

    __call__ = write


#------------------------------------------------------------------------------


class StackInfoStorage(FallbackStorage):
    """
    Message storage like LegacyFallbackStorage, except, every message
    would have a stack info, witch is helpful, for debugging.
    
    Stack info would only be added, if...
        ...settings.DEBUG == True
    or
        ...settings.MESSAGE_DEBUG == True
    
    Put this into your settings:
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    MESSAGE_STORAGE = "django_tools.utils.messages.StackInfoStorage"
    
    Template e.g.:
    ~~~~~~~~~~~~~~
    <ul class="messages">
    {% for message in messages %}
    <li{% if message.tags %} class="{{ message.tags }}"{% endif %}{% if message.stack_info %} title="click for stack info" onclick="alert('Stack info (limit: {{ message.stack_limit }}, url: [{{ message.full_path }}]):\n\n{{ message.stack_info }}');"{% endif %}>
       {{ message }}
    </li>
    {% endfor %}
    </ul>
    """
    _add_stackinfo = settings.DEBUG or getattr(settings, "MESSAGE_DEBUG", False)

    def add(self, *args, **kwargs):
        """ Add stackinfo to the message """
        super(StackInfoStorage, self).add(*args, **kwargs)

        if self._add_stackinfo:
            # info: self._queued_messages is a normal list, defined in BaseStorage()
            message_list = self._queued_messages
            try:
                last_message = message_list[-1]
            except IndexError:
                return

            last_message.full_path = self.request.get_full_path()
            last_message.stack_limit = STACK_LIMIT

            stack_info = get_stack_info(filepath_filter="django_tools", stack_limit=STACK_LIMIT, max_filepath_len=MAX_FILEPATH_LEN)
            stack_info_safe = "\\n".join([l.replace("\n", "\\n").replace("'", "&#x27;") for l in stack_info])
            last_message.stack_info = stack_info_safe


#------------------------------------------------------------------------------


def failsafe_message(msg, level=messages.INFO):
    """
    Display a message to the user.
    Use ThreadLocalMiddleware to get the current request object.
    If no request object is available, create a warning.
    """
    request = ThreadLocal.get_current_request()
    if request:
        # create a normal user message
        try:
            messages.add_message(request, level, msg)
        except Exception as err:
            # e.g.:
            # Without the django.contrib.messages middleware,
            # messages can only be added to authenticated users.
            msg += " (Error create a message: %s)" % err
        else:
            return

    # fallback: Create a warning
    warnings.warn(msg)



# coding: utf-8

"""
    test django_tools.template.loader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function

from django.template import TemplateDoesNotExist

from django_tools.unittest_utils.unittest_base import BaseTestCase


class DebugCacheLoaderTest(BaseTestCase):
    def test(self):
        response = self.client.get("/admin/login/")

        self.assertResponse(response,
            must_contain=(
                "<!-- START 'admin/login.html' -->",
                "<!-- START 'admin/base_site.html' -->",
                "<!-- START 'admin/base.html' -->",

                "<!-- END 'admin/base.html' -->",
                "<!-- END 'admin/base_site.html' -->",
                "<!-- END 'admin/login.html' -->",
            ),
            must_not_contain=(
                "Error", "Traceback",
            ),
            status_code=200,
            template_name="admin/login.html",
            html=False,
            browser_traceback=True
        )

    def test_template_does_not_exists(self):
        with self.assertRaises(TemplateDoesNotExist) as cm:
            self.client.get("/raise_template_not_exists/")

        output = "\n".join(cm.exception.args)
        self.assertEqual("/template/does/not/exists.html", output)



#!/usr/bin/env python3

import os
import sys

from django.contrib.auth import get_user_model
from django.contrib.staticfiles.management.commands.runserver import Command as RunServerCommand
from django.core.management import call_command

print("sys.real_prefix:", getattr(sys, "real_prefix", "-"))
print("sys.prefix:", sys.prefix)



class Command(RunServerCommand):
    """
    Expand django.contrib.staticfiles runserver
    """
    help = "Setup test project and run django developer server"

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument("--fresh", action="store_true", dest="delete_first", default=False,
            help="Delete existing entries.")

    def verbose_call(self, command, *args, **kwargs):
        self.stderr.write("_"*79)
        self.stdout.write("Call %r with: %r %r" % (command, args, kwargs))
        call_command(command, *args, **kwargs)

    def handle(self, *args, **options):

        delete_first=options.get('delete_first')

        if "RUN_MAIN" not in os.environ:
            # RUN_MAIN added by auto reloader, see: django/utils/autoreload.py
            self.verbose_call("makemigrations") # helpfull for developming and add/change models ;)
            self.verbose_call("migrate")

            # django.contrib.staticfiles.management.commands.collectstatic.Command
            self.verbose_call("collectstatic", interactive=False, link=True)

            User=get_user_model()
            qs = User.objects.filter(is_active = True, is_superuser=True)
            if qs.count() == 0:
                self.verbose_call("createsuperuser")

        options["insecure_serving"] = True
        super(Command, self).handle(*args, **options)



# coding: utf-8

"""
    test Mockups
    ~~~~~~~~~~~~
"""


import warnings

# https://github.com/jedie/django-tools
from django_tools.unittest_utils.mockup import ImageDummy
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin


class TestMockupImage(TestUserMixin, BaseTestCase):
    def test_create_pil_image(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # trigger all warnings

            pil_image = ImageDummy(width=300, height=150).create_pil_image()

            self.assertEqual(pil_image.width, 300)
            self.assertEqual(pil_image.height, 150)

            self.assertEqual(pil_image.verify(), None)

            self.assertEqual(len(w), 0)  # No warnings created

    def test_create_info_image(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # trigger all warnings

            image_dummy = ImageDummy(width=400, height=200)
            self.assertEqual(image_dummy.text_color, "#ffffff")
            self.assertEqual(image_dummy.text_align, "center")
            pil_image = image_dummy.create_info_image(text="foo")

            self.assertEqual(pil_image.width, 400)
            self.assertEqual(pil_image.height, 200)

            self.assertEqual(pil_image.verify(), None)

            self.assertEqual(len(w), 0)  # No warnings created

    def test_create_temp_filer_info_image(self):
        user = self.login(usertype="normal")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # trigger all warnings

            filer_info_image = ImageDummy(width=500, height=300).create_temp_filer_info_image(
                text="A test image", user=user
            )
            self.assertEqual(filer_info_image.width, 500)
            self.assertEqual(filer_info_image.height, 300)

            self.assertEqual(filer_info_image.size, 1791)
            self.assertIn(".png", filer_info_image.path)

            path = filer_info_image.path
            path = path.replace("-", "_")  # e.g.: /django-tools/ -> /django_tools/

            self.assertIn("/django_tools/django_tools_test_project/media/filer_public/", path)

            self.assertEqual(len(w), 0)  # No warnings created



# coding: utf-8


from __future__ import print_function, unicode_literals

import logging
import pprint

from django.contrib.auth.models import Group, Permission, User
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.test import override_settings

from django_tools_test_project.django_tools_test_app.models import LimitToUsergroupsTestModel, PermissionTestModel

# https://github.com/jedie/django-tools
from django_tools.permissions import (
    add_app_permissions, add_permissions, check_permission, get_admin_permissions, get_filtered_permissions,
    get_permission_by_string, has_perm, log_group_permissions, log_user_permissions, permissions2list,
    pprint_filtered_permissions
)
from django_tools.unittest_utils.logging_utils import LoggingBuffer
from django_tools.unittest_utils.stdout_redirect import StdoutStderrBuffer
from django_tools.unittest_utils.unittest_base import BaseTestCase
from django_tools.unittest_utils.user import TestUserMixin

log = logging.getLogger(__name__)


@override_settings(DEBUG = True)
class TestPermissions(TestUserMixin, BaseTestCase):
    @classmethod
    def setUpTestData(cls):
        super(TestPermissions, cls).setUpTestData()

        cls.extra_permission = get_permission_by_string(
            permission="django_tools_test_app.extra_permission"
        )

        cls.instance = PermissionTestModel.objects.create(foo="bar")

    def setUp(self):
        super(TestPermissions, self).setUp()

        users = User.objects.all()
        usernames = users.values_list("username", flat=True).order_by("username")
        assert tuple(usernames) == ('normal_test_user', 'staff_test_user', 'superuser')

        self.superuser = self._get_user(usertype="superuser")

        # The "staff user" has some permissions:
        self.staff_user = self._get_user(usertype="staff")
        self.staff_group = Group.objects.create(name="Staff User Group")
        self.staff_group.permissions.add(self.extra_permission)
        self.staff_group.permissions.add(get_permission_by_string("django_tools_test_app.add_permissiontestmodel"))
        self.staff_group.permissions.add(get_permission_by_string("django_tools_test_app.change_permissiontestmodel"))
        self.staff_group.permissions.add(get_permission_by_string("django_tools_test_app.delete_permissiontestmodel"))
        self.staff_user.groups.add(self.staff_group)
        self.staff_user.save()

        # The "normal user" has no permissions:
        self.normal_user = self._get_user(usertype="normal")
        self.normal_group = Group.objects.create(name="Normal User Group")
        self.normal_user.groups.add(self.normal_group)
        self.normal_user.save()

    #-------------------------------------------------------------------------

    def assert_permissiontestmodel(self, permission):
        self.assertIsInstance(permission, Permission)
        self.assertIsInstance(permission.content_type, ContentType)

        self.assertEqual(permission.content_type.app_label, "django_tools_test_app")
        self.assertEqual(permission.content_type.model, "permissiontestmodel")

    def assert_extra_permission(self, permission):
        self.assert_permissiontestmodel(permission)
        self.assertEqual(permission.codename, "extra_permission")

    #-------------------------------------------------------------------------

    def test_setup(self):
        self.assert_extra_permission(self.extra_permission)

        # The "normal user" has no permissions
        self.assertEqual(self.normal_user.get_all_permissions(), set())

        # The "staff user" has some permissions
        self.assertEqual(self.staff_user.get_all_permissions(),
            {
                "django_tools_test_app.extra_permission",
                "django_tools_test_app.add_permissiontestmodel",
                "django_tools_test_app.change_permissiontestmodel",
                "django_tools_test_app.delete_permissiontestmodel",
            }
        )

        all_permissions = [
            "%s.%s" % (entry.content_type, entry.codename)
            for entry in Permission.objects.all()
        ]
        pprint.pprint(all_permissions)

        # Default mode permissions:
        self.assertIn("permission test model.add_permissiontestmodel", all_permissions)
        self.assertIn("permission test model.change_permissiontestmodel", all_permissions)
        self.assertIn("permission test model.delete_permissiontestmodel", all_permissions)

        # Own permission defined via Meta.permissions:
        self.assertIn("permission test model.extra_permission", all_permissions)

    #-------------------------------------------------------------------------

    def test_get_default_permissions(self):
        # https://docs.djangoproject.com/en/1.8/ref/models/options/#default-permissions
        permission = get_permission_by_string("django_tools_test_app.add_permissiontestmodel")
        self.assert_permissiontestmodel(permission)
        self.assertEqual(permission.codename, "add_permissiontestmodel")

        permission = get_permission_by_string("django_tools_test_app.change_permissiontestmodel")
        self.assert_permissiontestmodel(permission)
        self.assertEqual(permission.codename, "change_permissiontestmodel")

        permission = get_permission_by_string("django_tools_test_app.delete_permissiontestmodel")
        self.assert_permissiontestmodel(permission)
        self.assertEqual(permission.codename, "delete_permissiontestmodel")

    def test_get_extra_permission(self):
        # https://docs.djangoproject.com/en/1.8/ref/models/options/#permissions
        permission = get_permission_by_string("django_tools_test_app.extra_permission")
        self.assert_extra_permission(permission)

    #-------------------------------------------------------------------------

    def test_get_permission_wrong_format(self):
        with self.assertRaises(PermissionDenied) as cm:
            get_permission_by_string("no_dot_in_string")

        self.assert_exception_startswith(cm, "Wrong permission string format")

    def test_get_permission_wrong_app_label(self):
        with self.assertRaises(PermissionDenied) as cm:
            get_permission_by_string("wrong.foobar")

        self.assert_exception_startswith(cm,
            "App label 'wrong' from permission 'wrong.foobar' doesn't exists!"
        )

    def test_get_permission_wrong_codename(self):
        with self.assertRaises(PermissionDenied) as cm:
            get_permission_by_string("auth.wrong")

        self.assert_exception_startswith(cm,
            "Codename 'wrong' from permission 'auth.wrong' doesn't exists!"
        )

    #-------------------------------------------------------------------------

    def test_add_permissions(self):
        permissions = (
            (LimitToUsergroupsTestModel, "extra_permission"),
        )
        add_permissions(permission_obj=self.normal_group, permissions=permissions)

        user = self.refresh_user(self.normal_user)
        self.assertEqual(user.get_all_permissions(), {'django_tools_test_app.extra_permission'})

    def test_check_permission_error_without_exception(self):
        self.assertFalse(
            check_permission(
                self.normal_user,
                "django_tools_test_app.extra_permission",
                raise_exception=False
            )
        )

    def test_check_permission_error_with_exception(self):
        with self.assertRaises(PermissionDenied) as cm:
            check_permission(self.normal_user, "django_tools_test_app.extra_permission")

        self.assertFalse(hasattr(cm, "args")) # No error message

    def test_check_permission_existing(self):
        self.assertTrue(
            check_permission(
                self.staff_user,
                "django_tools_test_app.extra_permission",
            )
        )

    #-------------------------------------------------------------------------

    def test_check_permission_wrong_format(self):
        check_permission(self.superuser, "no_dot_in_string") # superuser can do everything ;)

        with self.assertRaises(PermissionDenied) as cm:
            check_permission(self.staff_user, "no_dot_in_string")

        self.assert_exception_startswith(cm, "Wrong permission string format")

    def test_check_permission_wrong_app_label(self):
        check_permission(self.superuser, "wrong.foobar") # superuser can do everything ;)

        with self.assertRaises(PermissionDenied) as cm:
            check_permission(self.staff_user, "wrong.foobar")

        self.assert_exception_startswith(cm,
            "App label 'wrong' from permission 'wrong.foobar' doesn't exists!"
        )

    def test_check_permission_wrong_codename(self):
        check_permission(self.superuser, "auth.wrong") # superuser can do everything ;)

        with self.assertRaises(PermissionDenied) as cm:
            check_permission(self.staff_user, "auth.wrong")

        self.assert_exception_startswith(cm,
            "Codename 'wrong' from permission 'auth.wrong' doesn't exists!"
        )

    def test_get_admin_permissions(self):
        permissions = get_admin_permissions()
        permissions = permissions2list(permissions)
        pprint.pprint(permissions)
        self.assertEqual(permissions, [
            'auth.add_group',
            'auth.change_group',
            'auth.delete_group',
            'auth.add_user',
            'auth.change_user',
            'auth.delete_user',
            'django_tools_test_app.add_permissiontestmodel',
            'django_tools_test_app.change_permissiontestmodel',
            'django_tools_test_app.delete_permissiontestmodel',
            'django_tools_test_app.extra_permission',
            'filer.add_clipboard',
            'filer.change_clipboard',
            'filer.delete_clipboard',
            'filer.add_file',
            'filer.change_file',
            'filer.delete_file',
            'filer.add_folder',
            'filer.can_use_directory_listing',
            'filer.change_folder',
            'filer.delete_folder',
            'filer.add_folderpermission',
            'filer.change_folderpermission',
            'filer.delete_folderpermission',
            'filer.add_image',
            'filer.change_image',
            'filer.delete_image',
            'filer.add_thumbnailoption',
            'filer.change_thumbnailoption',
            'filer.delete_thumbnailoption',
            'flatpages.add_flatpage',
            'flatpages.change_flatpage',
            'flatpages.delete_flatpage',
            'sites.add_site',
            'sites.change_site',
            'sites.delete_site'
        ])

    def test_has_perm(self):
        self.assertTrue(has_perm(self.staff_user, "django_tools_test_app.change_permissiontestmodel"))

    def test_has_perm_log(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            self.assertEqual(has_perm(self.normal_user, "foo.bar1"), False)

        self.assertEqual(
            log.get_messages(),
            "DEBUG:django_tools.permissions:"
            "User normal_test_user has not foo.bar1"
        )

    def test_log_user_permissions1(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            log_user_permissions(self.normal_user)

        self.assertEqual(
            log.get_messages(),
            "DEBUG:django_tools.permissions:User 'normal_test_user' has no permission!"
        )

    def test_log_user_permissions2(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            log_user_permissions(self.staff_user)

        messages = [line.strip() for line in log.get_messages().splitlines()]
        pprint.pprint(messages)

        self.assertEqual(
            messages,
            [
                "DEBUG:django_tools.permissions:User 'staff_test_user' has permissions:",
                "* django_tools_test_app.add_permissiontestmodel",
                "* django_tools_test_app.change_permissiontestmodel",
                "* django_tools_test_app.delete_permissiontestmodel",
                "* django_tools_test_app.extra_permission"
            ]
        )

    def test_log_group_permissions1(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            log_group_permissions(self.normal_group)

        self.assertEqual(
            log.get_messages(),
            "DEBUG:django_tools.permissions:User group 'Normal User Group' has no permission!"
        )

    def test_log_group_permissions2(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            log_group_permissions(self.staff_group)

        messages = [line.strip() for line in log.get_messages().splitlines()]
        pprint.pprint(messages)

        self.assertEqual(
            messages,
            [
                "DEBUG:django_tools.permissions:User group 'Staff User Group' has permissions:",
                "* django_tools_test_app.add_permissiontestmodel",
                "* django_tools_test_app.change_permissiontestmodel",
                "* django_tools_test_app.delete_permissiontestmodel",
                "* django_tools_test_app.extra_permission"
            ]
        )

    def test_superuser_check(self):
        self.assertTrue(
            check_permission(
                self.superuser,
                permission="superuser check ignores this completely!",
            )
        )

    def test_has_no_extra_permission(self):
        self.assertFalse(
            self.instance.has_extra_permission_permission(
                user=self.normal_user,
                raise_exception=False
            )
        )

    def test_has_extra_permission(self):
        self.assertTrue(
            self.instance.has_extra_permission_permission(
                user=self.staff_user,
                raise_exception=False
            )
        )

    def test_has_default_permissions(self):
        self.assertTrue(self.instance.has_add_permission(user=self.staff_user))
        self.assertTrue(self.instance.has_change_permission(user=self.staff_user))
        self.assertTrue(self.instance.has_delete_permission(user=self.staff_user))

    #-------------------------------------------------------------------------

    def test_add_app_permissions(self):
        with LoggingBuffer(name=None, level=logging.ERROR) as log:
            add_app_permissions(
                permission_obj=self.normal_group,
                app_label="django_tools_test_app"
            )

        self.assertEqual(
            log.get_messages(),
            "DEBUG:django_tools.permissions:Add 10 permissions from app 'django_tools_test_app'"
        )

        permissions = self.normal_group.permissions.all()
        permissions = permissions2list(permissions)
        pprint.pprint(permissions)
        self.assertEqual(permissions, [
            'django_tools_test_app.add_limittousergroupstestmodel',
            'django_tools_test_app.change_limittousergroupstestmodel',
            'django_tools_test_app.delete_limittousergroupstestmodel',
            'django_tools_test_app.add_permissiontestmodel',
            'django_tools_test_app.change_permissiontestmodel',
            'django_tools_test_app.delete_permissiontestmodel',
            'django_tools_test_app.extra_permission',
            'django_tools_test_app.add_simpleparlermodel',
            'django_tools_test_app.change_simpleparlermodel',
            'django_tools_test_app.delete_simpleparlermodel'
        ])

    #-------------------------------------------------------------------------

    def test_get_filtered_permissions_without_any_filter(self):
        permissions = permissions2list(
            get_filtered_permissions() # without any filter -> we get all existing permissions
        )
        permissions.sort()
        pprint.pprint(permissions)

        all_permissions = permissions2list(Permission.objects.all())
        all_permissions.sort()

        self.assertEqual(permissions, all_permissions)

    def test_get_filtered_permissions(self):
        permissions = get_filtered_permissions(
            exclude_app_labels=("easy_thumbnails", "filer"),
            exclude_models=(LimitToUsergroupsTestModel, PermissionTestModel),
            exclude_codenames=("delete_group", "delete_user"),
            exclude_permissions=(
                (ContentType, "add_contenttype"),
                (ContentType, "delete_contenttype"),
            )
        )
        permissions = permissions2list(permissions)
        pprint.pprint(permissions)
        self.assertEqual(permissions, [
            'admin.add_logentry',
            'admin.change_logentry',
            'admin.delete_logentry',
            'auth.add_group',
            'auth.change_group',
            'auth.add_permission',
            'auth.change_permission',
            'auth.delete_permission',
            'auth.add_user',
            'auth.change_user',
            'contenttypes.change_contenttype',
            'django_tools_test_app.add_simpleparlermodel',
            'django_tools_test_app.change_simpleparlermodel',
            'django_tools_test_app.delete_simpleparlermodel',
            'flatpages.add_flatpage',
            'flatpages.change_flatpage',
            'flatpages.delete_flatpage',
            'sessions.add_session',
            'sessions.change_session',
            'sessions.delete_session',
            'sites.add_site',
            'sites.change_site',
            'sites.delete_site'
        ])

    def test_pprint_filtered_permissions_wrong_arguments(self):
        with self.assertRaises(AssertionError) as cm:
            pprint_filtered_permissions(["foo", "bar"])

        self.assert_exception_startswith(cm, "List must contain auth.models.Permission instances!")

    def test_pprint_filtered_permissions(self):
        permissions = get_filtered_permissions(
            exclude_app_labels=("easy_thumbnails", "filer"),
            exclude_actions=("delete",),
            exclude_models=(LimitToUsergroupsTestModel, PermissionTestModel),
            exclude_codenames=("change_group", "change_user"),
            exclude_permissions=(
                (ContentType, "add_contenttype"),
                (ContentType, "delete_contenttype"),
            )
        )
        with StdoutStderrBuffer() as buffer:
            pprint_filtered_permissions(permissions)

        output = buffer.get_output()
        print(output)
        self.assertEqual_dedent(output, """
            [*] admin.add_logentry
            [*] admin.change_logentry
            [ ] admin.delete_logentry
            [*] auth.add_group
            [ ] auth.change_group
            [ ] auth.delete_group
            [*] auth.add_permission
            [*] auth.change_permission
            [ ] auth.delete_permission
            [*] auth.add_user
            [ ] auth.change_user
            [ ] auth.delete_user
            [ ] contenttypes.add_contenttype
            [*] contenttypes.change_contenttype
            [ ] contenttypes.delete_contenttype
            [ ] django_tools_test_app.add_limittousergroupstestmodel
            [ ] django_tools_test_app.change_limittousergroupstestmodel
            [ ] django_tools_test_app.delete_limittousergroupstestmodel
            [ ] django_tools_test_app.add_permissiontestmodel
            [ ] django_tools_test_app.change_permissiontestmodel
            [ ] django_tools_test_app.delete_permissiontestmodel
            [ ] django_tools_test_app.extra_permission
            [*] django_tools_test_app.add_simpleparlermodel
            [*] django_tools_test_app.change_simpleparlermodel
            [ ] django_tools_test_app.delete_simpleparlermodel
            [ ] easy_thumbnails.add_source
            [ ] easy_thumbnails.change_source
            [ ] easy_thumbnails.delete_source
            [ ] easy_thumbnails.add_thumbnail
            [ ] easy_thumbnails.change_thumbnail
            [ ] easy_thumbnails.delete_thumbnail
            [ ] easy_thumbnails.add_thumbnaildimensions
            [ ] easy_thumbnails.change_thumbnaildimensions
            [ ] easy_thumbnails.delete_thumbnaildimensions
            [ ] filer.add_clipboard
            [ ] filer.change_clipboard
            [ ] filer.delete_clipboard
            [ ] filer.add_clipboarditem
            [ ] filer.change_clipboarditem
            [ ] filer.delete_clipboarditem
            [ ] filer.add_file
            [ ] filer.change_file
            [ ] filer.delete_file
            [ ] filer.add_folder
            [ ] filer.can_use_directory_listing
            [ ] filer.change_folder
            [ ] filer.delete_folder
            [ ] filer.add_folderpermission
            [ ] filer.change_folderpermission
            [ ] filer.delete_folderpermission
            [ ] filer.add_image
            [ ] filer.change_image
            [ ] filer.delete_image
            [ ] filer.add_thumbnailoption
            [ ] filer.change_thumbnailoption
            [ ] filer.delete_thumbnailoption
            [*] flatpages.add_flatpage
            [*] flatpages.change_flatpage
            [ ] flatpages.delete_flatpage
            [*] sessions.add_session
            [*] sessions.change_session
            [ ] sessions.delete_session
            [*] sites.add_site
            [*] sites.change_site
            [ ] sites.delete_site
        """)



# coding:utf-8

"""
    csrf related views
    ~~~~~~~~~~~~~~~~~~

    
    debug_csrf_failure()
    ~~~~~~~~~~~~~~~~~~~~
    Display the normal debug page and not the minimal csrf debug page.
    
    usage: Add this to your settings:
    -----------------------------------------------------------------------    
    CSRF_FAILURE_VIEW='django_tools.views.csrf.debug_csrf_failure'
    -----------------------------------------------------------------------


    :copyleft: 2012 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, division, print_function


from django.views.csrf import csrf_failure
from django.conf import settings
from django.http import Http404


class CsrfFailure(Exception):
    pass


def debug_csrf_failure(request, reason=""):
    """
    raised own CsrfFailure() exception to get the normal debug page on
    Csrf failures.
    See also:
        https://docs.djangoproject.com/en/1.3/ref/contrib/csrf/#rejected-requests
        
    More Info: See DocString above.
    """
    if not settings.DEBUG:
        # Use original HttpResponseForbidden:
        return csrf_failure(request, reason)

    raise CsrfFailure("csrf failure debug: %r" % reason)



"""
    :created: 13.06.2018 by Jens Diemer
    :copyleft: 2018 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import os
import tempfile
import unittest
from pathlib import Path

from django.conf import settings
from django.contrib import auth
from django.test import override_settings

from selenium.common.exceptions import NoSuchElementException

# https://github.com/jedie/django-tools
from django_tools.selenium.base import SeleniumBaseTestCase
from django_tools.selenium.chromedriver import SeleniumChromiumTestCase, chromium_available
from django_tools.selenium.django import (
    SeleniumChromiumStaticLiveServerTestCase, SeleniumFirefoxStaticLiveServerTestCase
)
from django_tools.selenium.geckodriver import SeleniumFirefoxTestCase, firefox_available
from django_tools.selenium.utils import find_executable
from django_tools.unittest_utils.user import TestUserMixin


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class ExampleChromiumTests(SeleniumChromiumStaticLiveServerTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title("Log in | Django site admin")
        self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class ExampleFirefoxTests(SeleniumFirefoxStaticLiveServerTestCase):
    def test_admin_login_page(self):
        self.driver.get(self.live_server_url + "/admin/login/")
        self.assert_equal_page_title("Log in | Django site admin")
        self.assert_in_page_source('<form action="/admin/login/" method="post" id="login-form">')
        self.assert_no_javascript_alert()


class SeleniumHelperTests(unittest.TestCase):
    def test_find_executable(self):
        with tempfile.NamedTemporaryFile(prefix="test_not_executable_", delete=False) as f:
            filepath = Path(f.name).resolve()
            self.assertTrue(filepath.is_file())
            name = filepath.name
            path = filepath.parent

            # File is not in PATH:
            with self.assertRaises(FileNotFoundError) as context_manager:
                find_executable(name)

            self.assertEqual(context_manager.exception.args, ("Can't find '%s' in PATH or None!" % name,))

            old_path = os.environ["PATH"]
            try:
                # File is in PATH, but not executable:
                os.environ["PATH"] += "%s%s" % (os.pathsep, path)
                with self.assertRaises(FileNotFoundError) as context_manager:
                    find_executable(name)

                self.assertEqual(context_manager.exception.args, ("%s exists, but it's not executable!" % filepath,))

                # File is in PATH and executable:
                filepath.chmod(0x777)
                result = find_executable(name)
                self.assertEqual(result, filepath)
            finally:
                os.environ["PATH"] = old_path

            # Executable file is not in PATH, but can be found via extra_search_paths:
            result = find_executable(name, extra_search_paths=(path,))
            self.assertEqual(result, filepath)


class SeleniumTestsMixin:
    def test_login(self):

        self.assertTrue(settings.DEBUG)

        staff_data = self.get_userdata("staff")

        login_button_xpath = '//input[@value="Log in"]'

        url = self.live_server_url + "/admin/login/?next=/admin/"
        self.driver.get(url)

        self.assert_in_page_source("Log in | Django site admin")
        self.assert_not_in_page_source("errornote")
        self.assert_not_in_page_source("Please enter the correct username and password")

        self.assert_visible_by_id("id_username", timeout=2)
        self.assert_clickable_by_id("id_password", timeout=2)
        self.assert_clickable_by_xpath(login_button_xpath, timeout=2)

        try:
            username_input = self.driver.find_element_by_name("username")
            username_input.send_keys(staff_data["username"])

            password_input = self.driver.find_element_by_name("password")
            password_input.send_keys(staff_data["password"])

            login_button = self.driver.find_element_by_xpath(login_button_xpath)
            login_button.click()
        except NoSuchElementException as ex:
            self.fail(ex.msg)

        self.assert_no_javascript_alert()
        self.assert_not_in_page_source("Please enter the correct username and password")
        self.assert_not_in_page_source("errornote")
        self.assert_equal_page_title("Site administration | Django site admin")
        self.assert_in_page_source("<strong>staff_test_user</strong>")

        self.assert_visible_by_id("user-tools", timeout=2)
        self.assert_clickable_by_id("site-name", timeout=2)

    def test_admin_static_files(self):
        self.driver.get(self.live_server_url + "/admin/login/?next=/admin/")
        self.assert_in_page_source('href="/static/admin/css/base.css"')

        self.driver.get(self.live_server_url + "/static/admin/css/base.css")
        self.assert_in_page_source("margin: 0;")
        self.assert_in_page_source("padding: 0;")

    def test_local_storage_access(self):
        self.driver.get(self.live_server_url + "/")

        self.assertEqual(len(self.local_storage), 0)

        with self.assertRaises(KeyError) as cm:
            self.assert_local_storage_key_value(key="foo", value="bar")
        self.assertEqual(cm.exception.args[0], "foo")

        self.local_storage["bar"] = "foo"
        self.assertEqual(len(self.local_storage), 1)
        self.assert_local_storage_key_value(key="bar", value="foo")

        self.assertIn("bar", self.local_storage)

        self.assertEqual(repr(self.local_storage), "{'bar': 'foo'}")

        # There's no type conversion!
        self.local_storage["one"] = 1
        self.local_storage["t"] = True
        self.local_storage["f"] = False

        self.assertEqual(self.local_storage.items(), {"bar": "foo", "f": "false", "t": "true", "one": "1"})
        self.local_storage.clear()

        self.assertEqual(self.local_storage.items(), {})

        self.local_storage["foo"] = "bar"
        self.assertEqual(self.local_storage.items(), {"foo": "bar"})

        self.tearDown() # tearDown should clear the local_storage
        self.assertEqual(self.local_storage.items(), {})


@override_settings(DEBUG=True)
@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class SeleniumChromiumAdminTests(TestUserMixin, SeleniumChromiumStaticLiveServerTestCase, SeleniumTestsMixin):
    def test_console(self):
        self.driver.get(self.live_server_url + "/")

        self.driver.execute_script("console.log('test console output 1');")
        self.assert_in_browser_log("test console output 1")

        self.driver.execute_script("console.log('test console output 2');")
        self.assert_in_browser_log("test console output 2")


@override_settings(DEBUG=True)
@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class SeleniumFirefoxAdminTests(TestUserMixin, SeleniumFirefoxStaticLiveServerTestCase, SeleniumTestsMixin):
    pass


class NonStaticLiveServerMixin:
    def test_google(self):
        self.driver.get("https://www.google.com/")
        self.assert_equal_page_title("Google")
        self.assert_in_page_source("Google Inc.")


@unittest.skipUnless(chromium_available(), "Skip because Chromium is not available!")
class TestSeleniumChromiumNonStaticLiveServer(SeleniumChromiumTestCase, NonStaticLiveServerMixin):
    pass


@unittest.skipUnless(firefox_available(), "Skip because Firefox is not available!")
class TestSeleniumFirefoxNonStaticLiveServer(SeleniumFirefoxTestCase, NonStaticLiveServerMixin):
    pass



# coding:utf-8

"""
    :copyleft: 2017 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import unicode_literals, absolute_import, print_function

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory


def create_fake_request(url="/", session=None, language_code=None, user=None, **extra):
    if session is None:
        session = {}

    if language_code is None:
        language_code = settings.LANGUAGE_CODE

    if user is None:
        user = AnonymousUser()

    fake_request = RequestFactory().get(url)

    fake_request.session = session
    fake_request.LANGUAGE_CODE = language_code
    fake_request.user = user

    for key, value in extra.items():
        setattr(fake_request, key, value)

    return fake_request



# coding: utf-8

"""
    DjangoCommandMixin
    ~~~~~~~~~~~~~~~~~~

    :copyleft: 2012-2015 by the django-tools team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import unicode_literals, print_function

import os
import pprint
import subprocess

import sys


class DjangoCommandMixin(object):
    def subprocess_getstatusoutput(self, cmd, debug=False, **kwargs):
        """
        Return (status, output) of executing cmd in a shell.

        similar to subprocess.getstatusoutput but pass though kwargs
        """

        # Assume that DJANGO_SETTINGS_MODULE not in environment
        # e.g:
        #   manage.py use os.environ.setdefault("DJANGO_SETTINGS_MODULE",...)
        #   so it will ignore the own module path!
        # You can set env by given kwargs, too.
        env=dict(os.environ)
        if "DJANGO_SETTINGS_MODULE" in env:
            del(env["DJANGO_SETTINGS_MODULE"])

        subprocess_kwargs = {
            "env": env,
            "shell": True,
            "universal_newlines": True,
            "stderr": subprocess.STDOUT,
        }
        subprocess_kwargs.update(kwargs)
        if "cwd" in subprocess_kwargs:
            cwd = subprocess_kwargs["cwd"]
            assert os.path.isdir(cwd), "cwd %r doesn't exists!" % cwd
            if debug:
                print("DEBUG: cwd %r, ok" % cwd)

        cmd=" ".join(cmd) # FIXME: Why?!?
        try:
            output = subprocess.check_output(cmd, **subprocess_kwargs)
            status = 0
        except subprocess.CalledProcessError as ex:
            output = ex.output
            status = ex.returncode

        if output[-1:] == '\n':
            output = output[:-1]

        if status != 0 or debug:
            msg = (
                "subprocess exist status == %(status)r\n"
                "Call %(cmd)r with:\n"
                "%(kwargs)s\n"
                "subprocess output:\n"
                "------------------------------------------------------------\n"
                "%(output)s\n"
                "------------------------------------------------------------\n"
            ) % {
                "status": status,
                "cmd": cmd,
                "kwargs": pprint.pformat(subprocess_kwargs),
                "output": output
            }
            if status != 0:
                raise AssertionError(msg)
            else:
                print(msg)

        return output

    def call_manage_py(self, cmd, manage_dir, manage_py="manage.py", assert_executable=True, **kwargs):
        """
        call manage.py from given >manage_dir<
        """
        test_path = os.path.join(manage_dir, manage_py)
        if not os.path.isfile(test_path):
            msg = (
                "File doesn't exists: %r"
                " (given <manage_dir> path wrong?!?)"
            ) % manage_dir
            raise AssertionError(msg)

        if assert_executable and not os.access(test_path, os.X_OK):
            msg = (
                "Manage file %r is not executable!"
            ) % test_path
            raise AssertionError(msg)

        cmd = [sys.executable, manage_py] + list(cmd)
        kwargs.update({
            "cwd": manage_dir,
            #"debug": True,
        })
        return self.subprocess_getstatusoutput(cmd, **kwargs)

